#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Re-Ranker Object.

Provided with a predictor model file, the re-ranker provides an API for re-ranking
candidate outputs.
"""
import logging
import torch
from abc import ABC, abstractmethod, abstractclassmethod
from typing import List, Optional, Tuple
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.agents import create_agent_from_model_file, Agent
from parlai.core.build_data import modelzoo_path
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import TorchAgent
from parlai.utils.strings import normalize_reply
from parlai.utils.torch import argsort
from parlai.agents.hugging_face.gpt2 import Gpt2Agent
from projects.msc.agents.long_tga import TransformerVariantAgent

RERANKER_STRATEGIES = ['sum_scores', 'hard_choice', 'reranker_score', 'none']


class AbstractReranker(ABC):
    """
    Re-ranker object which uses a predictor model to rerank candidates given.

    This AbstractReranker can be subclassed to use any classifier or ranking model as
    the predictor.
    """

    @classmethod
    def add_cmdline_args(cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None):
        reranker = parser.add_argument_group('AbstractReranker Args')
        reranker.add_argument(
            '--normalize-candidates',
            type=str,
            default=False,
            help='Remove spaces and add capitalization as per ParlAI normalize_reply() function',
        )
        reranker.add_argument(
            '--predictor-model-file',
            type=str,
            default=None,
            help='Path to model whose prediction score will be used to rerank, usually a classifier or ranker',
        )
        reranker.add_argument(
            '--reranker-strategy',
            type=str,
            default='reranker_score',
            help='Which strategy to use when re-ranking response candidates. '
            f"Choices: {','.join(RERANKER_STRATEGIES)}",
        )
        reranker.add_argument(
            '--reranker-delimiter',
            type=str,
            default=None,
            help='delimiter for the reranker',
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        """
        Initializes reranker.
        """
        self.predictor_model_file = modelzoo_path(
            opt['datapath'], opt['predictor_model_file']
        )
        self.reranker_strategy = opt['reranker_strategy']
        self.normalize_candidates = opt['normalize_candidates']
        self.delimiter = opt.get('reranker_delimiter', None)
        if not self.delimiter:
            self.delimiter = opt.get('delimiter', '\n')
        self.include_context = True
        self.include_label_cand_only = False
        self.init_predictor(opt, shared)

    def init_predictor(self, opt: Opt, shared=None):
        """
        Initializes Predictor Module.
        """
        if not shared:
            if not opt.get("predictor_model_file"):
                logging.warn(
                    'Reranker MUST specify predictor_model_file unless subclass __init__() sets up the model in its own way (unusual). Skipping predictor setup!'
                )
            else:
                self.predictor = create_agent_from_model_file(self.predictor_model_file)
        else:
            self.predictor = shared['predictor']

    def share(self):
        shared = {}
        shared['predictor'] = self.predictor
        return shared

    def get_utterances_from_full_context(
        self, full_context: str, include_context: Optional[bool] = True
    ) -> List[str]:
        """
        Return utterances to consider from context string.

        :param full_context:
            full context (or dialog history) to consider
        :param include_context:
            whether or not to include the context utterances (like "your persona:...")

        :return utterances:
            return list of utterances
        """
        split_context = full_context.split(self.delimiter)
        context = [l for l in split_context if self.is_context(l)]
        dialogue = [l for l in split_context if not self.is_context(l)]

        if include_context:
            utterances = context + dialogue
        else:
            utterances = dialogue
        return utterances

    def augment_context(
        self, full_context: str, candidate: str, include_context: bool
    ) -> List[str]:
        """
        Given context and candidate, augment the context with the new candidate.

        This is generally the input for the predictor model.

        :param full_context:
            dialogue context:
        :param candidate:
            candidate response
        :param include_context:
            Whether to include the context utterances (as defined by self.is_context())

        :return augmented_context:
            return the augmented context.
        """
        utterances = self.get_utterances_from_full_context(
            full_context, include_context=include_context
        )

        utterances.append(candidate)
        return utterances

    def predict(self, context: str, predictor_label_candidates: List[str]) -> Message:
        """
        Use predictor to predict given augmented context.

        :param context:
            augmented context with response candidates
        :param predictor_label_candidates:
            optional array of label candidates to pass to the predictor

        :return output:
            return output from ranker act
        """
        assert isinstance(self.predictor, Agent)
        obs = Message({'text': context, 'episode_done': True})
        obs['label_candidates'] = predictor_label_candidates
        self.predictor.observe(obs)
        act = self.predictor.act()
        assert isinstance(act, Message)
        return act

    def batch_predict(
        self, contexts: List[str], predictor_label_candidates: List[str]
    ) -> List[Message]:
        """
        Batch operate on a list of contexts.

        :param contexts:
            list of augmented contexts with response candidates
        :param predictor_label_candidates:
            array of label candidates to pass to the predictor
            (currently has to be the same for all the contexts)
        :return outputs:
            return ranker batch replies
        """
        assert isinstance(self.predictor, TorchAgent)
        observations = [{'text': context, 'episode_done': True} for context in contexts]
        processed_obs = []
        for obs in observations:
            obs['label_candidates'] = predictor_label_candidates
            obs = self.predictor.observe(obs)
            self.predictor.self_observe(obs)
            processed_obs.append(obs)
        batch_replies = self.predictor.batch_act(processed_obs)
        return batch_replies

    def _rerank_candidates(
        self,
        reranker_outputs: List[Message],
        response_cands: List[str],
        response_cand_scores: torch.Tensor,
        rerank_for_class: str,
    ) -> Tuple[List[str], List[int]]:
        """
        Re-rank the response candidates given reranker outputs and a strategy.

        Compute reranking differently according to specified rerank strategy

        :param reranker_outputs:
            outputs from reranker
        :param response_cands:
            list of response candidates
        :param response_cand_scores:
            tensor with scored response candidates from initial model
        :param rerank_for_class:
            The class (in the ML sense) we want to select for

        :return (candidates, indices):
            candidates: reranked list of response candidates.
            indices: list of indices into response_cands corresponding to re-rank order
        """
        if self.reranker_strategy == 'hard_choice':
            predicted_class = [
                (i, c)
                for i, c in enumerate(response_cands)
                if reranker_outputs[i]['text'] == rerank_for_class
            ]
            try:
                predicted_indices, predicted_class = [
                    list(l) for l in zip(*predicted_class)
                ]
            except ValueError:
                # none predicted
                predicted_indices, predicted_class = [], []

            predicted_not_class = [
                (i, c)
                for i, c in enumerate(response_cands)
                if reranker_outputs[i]['text'] != rerank_for_class
            ]
            try:
                predicted_not_indices, predicted_not_class = [
                    list(l) for l in zip(*predicted_not_class)
                ]
            except ValueError:
                predicted_not_indices, predicted_not_class = [], []

            candidates = predicted_class + predicted_not_class
            indices = predicted_indices + predicted_not_indices
        elif self.reranker_strategy == 'sum_scores':
            rerank_scores = [
                o['sorted_scores'][o['text_candidates'].index(rerank_for_class)]
                for o in reranker_outputs
            ]
            scores = [
                rerank_scores[i].item() + response_cand_scores[i].item()
                for i in range(len(rerank_scores))
            ]
            candidates, indices = argsort(
                scores,
                response_cands,  # type: ignore
                list(range(len(response_cands))),  # type: ignore
                descending=True,
            )[:2]
        elif self.reranker_strategy == 'reranker_score':
            rerank_scores = [
                o['sorted_scores'][o['text_candidates'].index(rerank_for_class)].item()
                for o in reranker_outputs
            ]
            candidates, indices = argsort(
                rerank_scores,
                response_cands,  # type: ignore
                list(range(len(response_cands))),  # type: ignore
                descending=True,
            )[:2]
        elif self.reranker_strategy == 'none':
            candidates = response_cands
            indices = list(range(len(response_cands)))

        return candidates, indices  # type: ignore

    def rerank(
        self,
        observation: Message,
        response_cands: List[str],
        response_cand_scores: torch.Tensor,
    ) -> Tuple[List[str], List[int]]:
        """
        Re-rank candidates according to predictor score.

        :param observation:
            Message object that includes the dialogue history
        :param response_cands:
            ranked list of model response candidates
        :param response_cand_scores:
            list of model response candidates' scores

        :return (candidates, indices):
            candidates: a re-ranked list of candidates
            indices: list of indices into response_cands corresponding to re-rank order
        """
        full_context = observation['full_text']

        # 0) Normalize the replies if the opt is passed in
        if self.normalize_candidates:
            response_cands = [normalize_reply(c) for c in response_cands]

        # 1) Augment context with response candidates
        if not self.include_label_cand_only:
            contexts = [
                self.augment_context(
                    full_context, cand, include_context=self.include_context
                )
                for cand in response_cands
            ]
            contexts = [self.delimiter.join(utts) for utts in contexts]
        else:
            # This variant only passes in the label candidates (with no dialogue history whatsoever
            # into the ranker. Can be useful for things like e.g. simple utterance-based safety classifiers.
            contexts = response_cands

        # 2) Predict with augmented context
        label_candidates = self.get_predictor_label_candidates(
            observation, full_context
        )
        reranker_outputs = self.batch_predict(contexts, label_candidates)

        # 3) Rerank
        rerank_for_class = self.get_class_to_rerank_for(observation, full_context)
        reranked_candidates, indices = self._rerank_candidates(
            reranker_outputs,
            response_cands,
            response_cand_scores,
            rerank_for_class=rerank_for_class,
        )

        if self.show_debug_logging(observation):
            debug_str = self._construct_debug_logging(
                observation,
                response_cands,
                response_cand_scores,
                reranker_outputs,
                reranked_candidates,
                rerank_for_class,
            )
            # Need print because even logging.WARN swallowed during eval
            print(debug_str)
        return reranked_candidates, indices

    ###################
    # DEBUG CODE ######
    ###################

    def show_debug_logging(self, observation: Message) -> bool:
        """
        Determine whether to show debug logging.

        Override to perform some logic to determine if this is the case.

        :param observation:
            observation under consideration

        :return show_debug:
            return true if we should show debug logging.
        """
        return False

    def _construct_debug_logging(
        self,
        observation: Message,
        response_cands: List[str],
        response_cand_scores: torch.Tensor,
        reranker_outputs: List[Message],
        reranked_candidates: List[str],
        rerank_for_class: str,
    ) -> str:
        """
        Construct debug logging.
        """
        s = '-------------\nFull Context:\n'
        full_context = observation['full_text']
        utterances = full_context.split(self.delimiter)
        count = 0
        for utt in utterances:
            if self.is_context(utt):
                s += utt
            else:
                s += f'{"Human: " if count % 2 == 0 else "Model: "} {utt}\n'
                count += 1

        s += f'Label was: {observation["eval_labels"]}\n'
        s += 'Before Reranking:\n'
        for i, c in enumerate(response_cands):
            rerank_score = reranker_outputs[i]['sorted_scores'][
                reranker_outputs[i]['text_candidates'].index(rerank_for_class)
            ]
            s += f'c: {c}, beam score: {response_cand_scores[i].item()}, rerank score: {rerank_score}\n'

        s += 'Reranked candidates:\n'
        for c in reranked_candidates:
            s += f'c: {c}\n'
        return s

    ########################
    # Abstract Methods #####
    ########################

    @abstractmethod
    def get_class_to_rerank_for(self, observation: Message, full_context: str) -> str:
        """
        The class from the predictor (classifier) that we will rerank to favor.

        This must correspond to an output from the predictor.
        """

    @abstractmethod
    def is_context(self, utt: str) -> bool:
        """
        Determine if incoming utterance is part of the context.

        :param utt:
            an utterance

        :return is_context:
            returns whether the incoming utterance is part of the context
        """

    @abstractmethod
    def get_predictor_label_candidates(
        self, observation: Message, context: str
    ) -> List[str]:
        """
        Get the list of possible predictor classes.

        Subclasses must override to return possible classes, given a context string.
        """


class AbstractGeneratorRerankAgentMixin:
    """
    Generator Rerank Agent.

    Utilize a classifier/predictor model as re-ranker with a generative agent.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        AbstractReranker.add_cmdline_args(parser)
        parser.set_defaults(skip_generation=False)
        gen_agent = parser.add_argument_group("Generator Rerank Agent")
        gen_agent.add_argument(
            '--inference-strategies',
            type=str,
            default=None,
            help='comma-separated list of inference strategies. '
            'if specified, re-rank over several inference strategies',
        )
        gen_agent.add_argument(
            '--debug-mode',
            type='bool',
            default=False,
            help='specify to enable certain debugging procedures.',
        )
        gen_agent.add_argument(
            '--inference-opt-key',
            type=str,
            default='inference',
            help='specify inference opt key for dialogue response model',
        )

        return parser

    def __init__(self, opt: Opt, shared=None):
        """
        Setup reranker.
        """
        super().__init__(opt, shared)
        reranker_class = self.get_reranker_class()
        self.inference_opt_key = opt.get('inference_opt_key', 'inference')
        self.inference_strategies = (
            opt['inference_strategies'] or opt[self.inference_opt_key]
        ).split(',')
        self.debug_mode = opt.get('debug_mode', False)
        if not shared:
            self.reranker = reranker_class(opt, shared=None)
        else:
            self.reranker = reranker_class(opt, shared=shared['reranker'])

    @abstractclassmethod
    def get_reranker_class(cls) -> AbstractReranker:
        """
        Return class to instantiate re-ranker.
        """

    def set_rerank_strategy(self, strategy: str):
        """
        Set new rerank strategy.
        """
        assert strategy in RERANKER_STRATEGIES
        self.reranker.reranker_strategy = strategy

    def share(self):
        """
        Share model parameters.

        :return: dict of shared class and opt
        """
        shared = super().share()
        shared['reranker'] = self.reranker.share()
        return shared

    def set_decoding_method(self, strategy):
        self.opt[self.inference_opt_key] = strategy

    def get_observations_for_reranker(
        self, observations: List[Message], batch_reply: List[Message]
    ) -> List[Message]:
        return observations

    def batch_act(self, observations: List[Message]) -> List[Message]:
        """
        Batch process a list of observations.

        We call batch act directly on the generator, and then individually process each
        batch reply.
        """
        batch_reply = [Message() for _ in range(len(observations))]
        # 1. get all beam texts to consider
        for strategy in self.inference_strategies:
            self.set_decoding_method(strategy)
            inference_batch_reply = super().batch_act(observations)
            for i, resp in enumerate(inference_batch_reply):
                beam_texts = batch_reply[i].get('beam_texts', [])
                batch_reply[i] = resp  # add metrics, other response items
                new_beam_texts = [(*b, strategy) for b in resp.get('beam_texts', [])]
                batch_reply[i].force_set('beam_texts', beam_texts + new_beam_texts)
        # 2. Rerank
        observations_for_reranker = self.get_observations_for_reranker(
            observations, batch_reply
        )
        for observation, generator_response in zip(
            observations_for_reranker, batch_reply
        ):
            if (
                'beam_texts' not in generator_response
                or not generator_response['beam_texts']
            ):
                logging.warn(
                    f'Generator response had no "beam_texts" field and was: {generator_response}. Skipping reranking. Something could be seriously wrong, but this also occurs when batchsize is not a multiple of number of examples'
                )
                continue
            reranked_candidates, indices = self.reranker.rerank(
                observation,
                [b[0] for b in generator_response['beam_texts']],  # text
                torch.tensor([b[1] for b in generator_response['beam_texts']]),  # score
            )
            if self.debug_mode:
                reranked_candidates = [
                    f"*{generator_response['beam_texts'][i][-1]}* {text}"
                    for i, text in zip(indices, reranked_candidates)
                ]
            generator_response.force_set('text', reranked_candidates[0])
            generator_response.force_set('beam_texts', reranked_candidates)
        return batch_reply


class AbstractGeneratorRerankAgent(
    AbstractGeneratorRerankAgentMixin, TransformerGeneratorAgent, ABC
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        TransformerGeneratorAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        AbstractGeneratorRerankAgentMixin.add_cmdline_args(parser, partial_opt)
        reranker_class = cls.get_reranker_class() or AbstractReranker
        reranker_class.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class LongAbstractGeneratorRerankAgent(
    AbstractGeneratorRerankAgentMixin, TransformerVariantAgent, ABC
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        TransformerVariantAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        AbstractGeneratorRerankAgentMixin.add_cmdline_args(parser, partial_opt)
        reranker_class = cls.get_reranker_class() or AbstractReranker
        reranker_class.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser


class AbstractGpt2RerankAgent(AbstractGeneratorRerankAgentMixin, Gpt2Agent, ABC):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        Gpt2Agent.add_cmdline_args(parser, partial_opt=partial_opt)
        AbstractGeneratorRerankAgentMixin.add_cmdline_args(parser, partial_opt)
        reranker_class = cls.get_reranker_class() or AbstractReranker
        reranker_class.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Script for making light modifications to conversations from tod chats such that they are
ready for the ACUTE format.

Notably, this does things that are slightly too much of a pain in the butt to do with
regexes like "add suffixes to ids when multiple ids might have the same string" (and
change metadata appropriately).

For example, the following

```
python cleanup_conversation.py --source_file <insert filepath>_conversations.jsonl --report-path <insert filepath>.json --agent-suffixes user_utt_model _BASE_USER system_utt_model _BASE_SYSTEM --included-speakers goal_grounding_model user_utt_model system_utt_model
```

strips the API call related turns and adds "_BASE_USER" and "_BASE_SYSTEM" (which otherwise would be the model type name, ex BART) to the latter two, respecitvely.
"""

from parlai.core.params import ParlaiParser
from parlai.utils.conversations import Conversations, Metadata
from parlai.utils.io import PathManager
from parlai.core.script import ParlaiScript, register_script

from parlai.core.tod.tod_core import TodAgentType, TOD_AGENT_TYPE_TO_PREFIX

import json


@register_script("conversation_cleanup")
class ConversationCleanup(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(
            False,
            False,
            "Script for simplying conversations output from TOD. Input expected to be in conversations format, as is output",
        )
        # Following params are same as the `eval_model` script
        parser.add_argument(
            "--source-file",
            type=str,
            required=True,
            help="Source file in conversations format, generated from `tod_world_script.py`",
        )
        parser.add_argument(
            "--out-file", type=str, default=None, help="Output location."
        )
        parser.add_argument(
            "--included-speakers",
            nargs="*",
            type=str,
            choices=[e.value for e in TodAgentType],
            default=[TodAgentType.USER_UTT_AGENT, TodAgentType.SYSTEM_UTT_AGENT],
            help="Which of the speakers to not remove. Should match those in `tod_world`",
        )
        parser.add_argument(
            "--agent-suffixes",
            nargs="*",
            type=str,
            default=[
                TodAgentType.USER_UTT_AGENT,
                "_USER",
                TodAgentType.SYSTEM_UTT_AGENT,
                "_SYSTEM",
            ],
            help="List of <speaker type, suffix> pairs. Speaker type should match those in `TodAgentType`; outputs (if included) will have the suffix added to the ID. This is useful when using multiple of the same out model (ex. Bart model for both the user and the system)",
        )
        parser.add_argument(
            "--num-conversations",
            default=400,
            help="Number of conversations to include. -1 for all",
        )
        parser.add_argument(
            "--report-path",
            required=True,
            help="path of the report saved from the tod_metrics_script",
        )
        return parser

    def _get_turn_type(self, turn):
        for agent_type, prefix in TOD_AGENT_TYPE_TO_PREFIX.items():
            if prefix in turn["text"]:
                return agent_type

    def run(self):
        opt = self.opt
        if int(len(self.opt["agent_suffixes"])) % 2 != 0:
            raise RuntimeError("Agent suffix input should be even")
        suffixes = {}
        for i in range(int(len(self.opt["agent_suffixes"]) / 2)):
            agent = self.opt["agent_suffixes"][2 * i]
            suffix = self.opt["agent_suffixes"][2 * i + 1]
            suffixes[agent] = suffix

        with PathManager.open(opt["report_path"]) as r:
            report = json.load(r)["report"]
        tod_metrics = report["tod_metrics"]

        if opt["num_conversations"] > -1:
            tod_metrics = tod_metrics[: opt["num_conversations"]]

        source = self.opt["source_file"].replace(".jsonl", "")
        if self.opt["out_file"]:
            out = self.opt["out_file"]
        else:
            if (
                "conversations" in source
            ):  # just to make sure we don't overwrite anything...
                out = source.replace("conversations", "cleaned_conversations")
            else:
                out = "cleaned_" + source

        speakers = []
        with PathManager.open(out + ".jsonl", "w") as f:
            conversations = Conversations(source + ".jsonl")
            for i, conversation in enumerate(conversations):
                if opt["num_conversations"] >= 0 and i >= opt["num_conversations"]:
                    break
                cleaned_dialog = []
                for parlay_round in conversation.episode["dialog"]:
                    cleaned_parlay_round = []
                    for turn in parlay_round:
                        turn_type = self._get_turn_type(turn)
                        if turn_type in self.opt["included_speakers"]:
                            if turn_type in suffixes:
                                turn["id"] += suffixes[turn_type]
                            if turn["id"] not in speakers:
                                speakers.append(turn["id"])
                            cleaned_parlay_round.append(turn)
                    if len(cleaned_parlay_round) > 0:
                        cleaned_dialog.append(cleaned_parlay_round)
                convo = {}
                convo["dialog"] = cleaned_dialog
                convo["metadata_path"] = Metadata._get_path(out)
                convo["context"] = [
                    {
                        "synthetic_task_success": tod_metrics[i][
                            "synthetic_task_success"
                        ],
                        "goal_text": tod_metrics[i]["goal"]["text"],
                    }
                ]
                json_convo = json.dumps(convo)
                f.write(json_convo + "\n")

            old_meta = Metadata(source + ".jsonl")
            Metadata.save_metadata(
                out, old_meta.opt, old_meta.self_chat, speakers, **old_meta.extra_data
            )


if __name__ == "__main__":
    ConversationCleanup.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import defaultdict
from typing import Dict, List, Optional, Tuple

import torch.jit
from parlai.agents.bart.bart import BartAgent
from parlai.core.dict import DictionaryAgent
from parlai.core.torch_agent import TorchAgent
from parlai.utils.bpe import Gpt2BpeHelper, SubwordBPEHelper
from parlai.torchscript.tokenizer import ScriptableDictionaryAgent
from torch import nn as nn


class TorchScriptGreedySearch(nn.Module):
    """
    A helper class for exporting simple greedy-search models via TorchScript.

    Models with extra inputs will need to override to include more variables.
    """

    # We currently only support these specific dictionary settings
    CAIRAOKE_DICT_PARAMS = {
        "dict_class": "parlai.core.dict:DictionaryAgent",
        "dict_initpath": None,
        "dict_language": "english",
        "dict_max_ngram_size": -1,
        "dict_minfreq": 0,
        "dict_maxtokens": -1,
        "dict_tokenizer": ["gpt2", "slow_bytelevel_bpe"],
        "dict_textfields": "text,labels",
        "dict_loaded": True,
        "bpe_debug": False,
    }

    def __init__(self, agent: TorchAgent):
        super().__init__()

        self.is_bart = isinstance(agent, BartAgent)
        self.device = agent.model.encoder.embeddings.weight.device
        # Dictionary/tokenization setup
        for key, val in self.CAIRAOKE_DICT_PARAMS.items():
            if type(val) == list and len(val) > 0:
                assert (
                    agent.opt.get(key, val[0]) in val
                ), f'The only currently supported values of "{key}" are {", ".join(val)}!'
            else:
                assert (
                    agent.opt.get(key, val) == val
                ), f'The only currently supported value of "{key}" is {val}!'
        orig_dict: DictionaryAgent = agent.dict
        orig_bpe: Gpt2BpeHelper = orig_dict.bpe
        assert all(len(key) == 2 for key in orig_bpe.bpe_ranks.keys())
        if any(i for key in orig_bpe.bpe_ranks.keys() for i in key if "\n" in i):
            raise AssertionError(
                "We need to temporarily merge the bpe_ranks dict's keys with a newline "
                "character in order to use it as a TorchScript arg, but at least one of the "
                "dict's keys contains a newline character already!"
            )
        fused_key_bpe_ranks = {
            "\n".join(key): float(val) for key, val in orig_bpe.bpe_ranks.items()
        }
        # Cast the values as floats to be able to compare to float('inf') when doing BPE
        # splitting
        self.dict = ScriptableDictionaryAgent(
            null_token=orig_dict.null_token,
            end_token=orig_dict.end_token,
            unk_token=orig_dict.unk_token,
            start_token=orig_dict.start_token,
            freq=orig_dict.freq,
            tok2ind=orig_dict.tok2ind,
            ind2tok=orig_dict.ind2tok,
            bpe_add_prefix_space=agent.opt["bpe_add_prefix_space"],
            bpe_encoder=orig_bpe.encoder,
            bpe_byte_encoder=orig_bpe.byte_encoder,
            fused_key_bpe_ranks=fused_key_bpe_ranks,
            special_tokens=agent._get_special_tokens(),
            subword_bpe_version=(0, 0),
            fused_bpe_codes={},
            subword_bpe_separator='',
        )

        # History tracking and start/end tokens
        self.delimiter_tok = agent.history.delimiter_tok
        self.history_size = agent.opt["history_size"]
        if agent.opt.get("history_add_global_end_token", None) is not None:
            self.global_end_token = agent.dict[agent.dict.end_token]
        else:
            self.global_end_token = None
        self.text_truncate = agent.opt.get("text_truncate") or agent.opt["truncate"]
        self.text_truncate = self.text_truncate if self.text_truncate >= 0 else None

        self.start_idx = agent.model.START_IDX
        self.end_idx = agent.model.END_IDX
        self.null_idx = agent.model.NULL_IDX
        if self.is_bart:
            self.initial_decoder_input = [self.end_idx, self.start_idx]
        else:
            self.initial_decoder_input = [self.start_idx]

        agent.model.eval()

        # Create versions of the model and decoder that will flatten the incremental
        # state dict, as required by TorchScript
        wrapped_decoder = DecoderIncrStateFlattener(agent.model.decoder)
        wrapped_model = ModelIncrStateFlattener(agent.model)

        # Create sample inputs for tracing
        sample_tokens = torch.tensor(
            [[1, 2, 3, 4, 5]], dtype=torch.long, device=self.device
        )
        sample_tokens = sample_tokens.to(self.device)
        encoder_states = agent.model.encoder(sample_tokens)
        initial_generations = self._get_initial_decoder_input(sample_tokens)
        latent, initial_incr_state = wrapped_decoder(
            initial_generations, encoder_states
        )
        logits = agent.model.output(latent[:, -1:, :])
        _, preds = logits.max(dim=2)
        incr_state = {k: torch.clone(v) for k, v in initial_incr_state.items()}
        # Copy the initial incremental state, used when tracing the
        # .reorder_decoder_incremental_state() method below, to avoid having it be
        # mutated by the following line
        incr_state = wrapped_model.reorder_decoder_incremental_state(
            incr_state, torch.tensor([0], dtype=torch.long, device=sample_tokens.device)
        )
        generations = torch.cat([initial_generations, preds], dim=1)

        # Do tracing
        self.encoder = torch.jit.trace(agent.model.encoder, sample_tokens)
        self.decoder_first_pass = torch.jit.trace(
            wrapped_decoder, (initial_generations, encoder_states), strict=False
        )
        # We do strict=False to avoid an error when passing a Dict out of
        # decoder.forward()
        self.partially_traced_model = torch.jit.trace_module(
            wrapped_model,
            {
                "output": (latent[:, -1:, :]),
                "reorder_decoder_incremental_state": (
                    initial_incr_state,
                    torch.tensor([0], dtype=torch.long, device=sample_tokens.device),
                ),
            },
            strict=False,
        )
        self.decoder_later_pass = torch.jit.trace(
            wrapped_decoder, (generations, encoder_states, incr_state), strict=False
        )

    def get_device(self):
        return self.encoder.embeddings.weight.device

    def _get_initial_decoder_input(self, x: torch.Tensor) -> torch.Tensor:
        """
        Workaround because we can't use TGM._get_initial_decoder_input() directly.

        When we try to call that function, we get a "RuntimeError: Type 'Tuple[int,
        int]' cannot be traced. Only Tensors and (possibly nested) Lists, Dicts, and
        Tuples of Tensors can be traced" error.
        """
        bsz = x.size(0)
        return (
            torch.tensor(
                self.initial_decoder_input, dtype=torch.long, device=self.device
            )
            .expand(bsz, len(self.initial_decoder_input))
            .to(x.device)
        )

    def parse(self, text: str) -> List[int]:
        return self.dict.txt2vec(text, dict_tokenizer='gpt2')

    def _v2t(self, vec: List[int]) -> str:
        """
        Convert token indices to string of tokens.
        """
        new_vec: List[int] = []
        for i in vec:
            if i == self.end_idx:
                break
            elif i != self.start_idx:
                new_vec.append(i)
        return self.dict.vec2txt(new_vec, dict_tokenizer='gpt2')

    def forward(self, context: str, max_len: int = 128) -> str:

        # Vectorize all lines of context
        history_vecs: List[List[int]] = []
        context_lines = context.split("\n")
        context_lines = self.preprocess_context(context_lines)
        if self.history_size > 0:
            context_lines = context_lines[-self.history_size :]
        for line in context_lines:
            history_vecs.append(self.parse(line))

        # Get full history vec
        text_vecs: List[List[int]] = []
        for vec in history_vecs[:-1]:
            text_vecs += [vec]
            text_vecs += [self.delimiter_tok]
        text_vecs += [history_vecs[-1]]
        if self.global_end_token is not None:
            text_vecs += [[self.global_end_token]]

        # Flatten text_vecs
        flattened_text_vec: List[int] = []
        for vec in text_vecs:
            for token in vec:
                flattened_text_vec.append(token)

        # Format history vec given various logic
        if self.text_truncate is not None:
            if self.is_bart:
                truncate_length = self.text_truncate - 2  # Start and end tokens
            else:
                truncate_length = self.text_truncate
            if len(flattened_text_vec) > truncate_length:
                flattened_text_vec = flattened_text_vec[-truncate_length:]
        flattened_text_vec = torch.tensor(flattened_text_vec, dtype=torch.long)
        if self.is_bart:
            flattened_text_vec = torch.cat(
                [
                    torch.tensor([self.start_idx], dtype=torch.long).to(
                        self.get_device()
                    ),
                    flattened_text_vec.to(self.get_device()),
                    torch.tensor([self.end_idx], dtype=torch.long).to(
                        self.get_device()
                    ),
                ],
                dim=0,
            )

        # Pass through the encoder and decoder to generate tokens

        flattened_text_vec = flattened_text_vec.to(self.get_device())
        batch_text_vec = torch.unsqueeze(flattened_text_vec, dim=0)  # Add batch dim
        encoder_states = self.encoder(batch_text_vec)
        generations = self._get_initial_decoder_input(batch_text_vec)
        # keep track of early stopping if all generations finish
        seen_end = torch.zeros(
            batch_text_vec.size(0), device=batch_text_vec.device, dtype=torch.bool
        )
        incr_state: Dict[str, torch.Tensor] = {}
        for token_idx in range(max_len):
            if token_idx == 0:
                latent, incr_state = self.decoder_first_pass(
                    generations, encoder_states
                )
            else:
                latent, incr_state = self.decoder_later_pass(
                    generations, encoder_states, incr_state
                )
            logits = self.partially_traced_model.output(latent[:, -1:, :])
            _, preds = logits.max(dim=2)
            incr_state = self.partially_traced_model.reorder_decoder_incremental_state(
                incr_state,
                torch.tensor([0], dtype=torch.long, device=batch_text_vec.device),
            )
            seen_end = seen_end + (preds == self.end_idx).squeeze(1)
            generations = torch.cat([generations, preds], dim=1)
            if torch.all(seen_end):
                break

        # Get the label from the generated tokens and update the history
        if self.is_bart:
            assert generations[0, 0].item() == self.end_idx
            generations = generations[:, 1:]
            # Hack: remove initial end token. I haven't found in the code where this is
            # done, but it seems to happen early on during generation
        generation_tokens: List[int] = generations[0].tolist()
        label = self._v2t(generation_tokens)

        return self.postprocess_output_generations(label=label)

    def postprocess_output_generations(self, label: str) -> str:
        """
        Post-process the model output.

        Returns the model output by default, override to add custom logic
        """
        return label

    def preprocess_context(self, context_lines: List[str]) -> List[str]:
        return context_lines


class BaseIncrStateFlattener(nn.Module):
    """
    Flatten/unflatten the incremental state for use with TorchScripting.

    Typically, the incremental state will be stored as a Dict[int, Dict[str, Dict[str,
    torch.Tensor]]], where the 3 dictionary levels map decoder layer, attention type,
    and previous key/value/mask, respectively. However, TorchScript expects dicts to be
    of type Dict[str, torch.Tensor], and thus all input incremental states when
    TorchScripting will have to be of that type. We thus unflatten the input incremental
    state, already of type Dict[str, torch.Tensor], to pass it into whatever method
    needs it, and we flatten it again after the updated incremental state is passed back
    out.

    This is a base class that provides methods for flattening/unflattening: subclasses
    will call these methods as the incremental state is passed into and out of their own
    methods.
    """

    def __init__(self, module: nn.Module):
        super().__init__()
        self.module = module

    def _unflatten_incr_state(
        self, flat_incr_state: Dict[str, torch.Tensor]
    ) -> Dict[int, Dict[str, Dict[str, torch.Tensor]]]:
        """
        Unflatten the input incremental state.

        For instance, flat_incr_state['layer_0__self_attn__prev_key'] will be stored in
        structured_incr_state[0]['self_attn']['prev_key'].
        """
        structured_incr_state = defaultdict(lambda: defaultdict(dict))
        for key, state in flat_incr_state.items():
            layer_idx_str, attn_type, state_type = key.split("__")
            structured_incr_state[int(layer_idx_str)][attn_type][state_type] = state
        return dict({k: dict(v) for k, v in structured_incr_state.items()})
        # Turn the nested defaultdicts back into regular dicts

    def _flatten_incr_state(
        self, structured_incr_state: Dict[int, Dict[str, Dict[str, torch.Tensor]]]
    ) -> Dict[str, torch.Tensor]:
        """
        Flatten the input incremental state.

        For instance, structured_incr_state[0]['self_attn']['prev_key'] will be stored
        in flat_incr_state['layer_0__self_attn__prev_key'].
        """
        flat_incr_state = {}
        for layer_idx, dict1 in structured_incr_state.items():
            for attn_type, dict2 in dict1.items():
                for state_type, state in dict2.items():
                    key = f"{layer_idx:d}__{attn_type}__{state_type}"
                    flat_incr_state[key] = state
        return flat_incr_state


class DecoderIncrStateFlattener(BaseIncrStateFlattener):
    """
    Wrapper for a TransformerDecoder that will unflatten/flatten the incremental state.

    Unflattening/flattening will occur before passing the incremental state into and out
    of .forward().
    """

    def forward(
        self,
        input_: torch.LongTensor,
        encoder_state: Tuple[torch.Tensor, torch.Tensor],
        flat_incr_state: Optional[Dict[str, torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, Dict[str, torch.Tensor]]:
        if flat_incr_state is not None:
            structured_incr_state = self._unflatten_incr_state(flat_incr_state)
        else:
            structured_incr_state = None
        tensor, new_structured_incr_state = self.module.forward(
            input=input_, encoder_state=encoder_state, incr_state=structured_incr_state
        )
        new_flat_incr_state = self._flatten_incr_state(new_structured_incr_state)
        return tensor, new_flat_incr_state


class ModelIncrStateFlattener(BaseIncrStateFlattener):
    """
    Wrapper for a TransformerGeneratorModel to unflatten/flatten the incremental state.

    Unflattening/flattening will occur before passing the incremental state into and out
    of .reorder_decoder_incremental_state(). We also support .output(), which is also
    traced.
    """

    def reorder_decoder_incremental_state(
        self, flat_incr_state: Dict[str, torch.Tensor], inds: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        structured_incr_state = self._unflatten_incr_state(flat_incr_state)
        new_structured_incr_state = self.module.reorder_decoder_incremental_state(
            incremental_state=structured_incr_state, inds=inds
        )
        return self._flatten_incr_state(new_structured_incr_state)

    def output(self, tensor: torch.Tensor) -> torch.Tensor:
        return self.module.output(tensor)


class TorchScriptTransformerClassifier(nn.Module):
    """
    A helper class for exporting transformer classifier via TorchScript.
    """

    def __init__(self, agent: TorchAgent):
        super().__init__()
        self.device = agent.model.transformer.embeddings.weight.device
        self.start_idx = agent.START_IDX
        self.end_idx = agent.END_IDX

        orig_dict: DictionaryAgent = agent.dict
        orig_bpe: SubwordBPEHelper = orig_dict.bpe

        assert all(len(key) == 2 for key in orig_bpe.bpe.bpe_codes.keys())
        if any(i for key in orig_bpe.bpe.bpe_codes.keys() for i in key if "\n" in i):
            raise AssertionError(
                "We need to temporarily merge the bpe_codes dict's keys with a newline "
                "character in order to use it as a TorchScript arg, but at least one of the "
                "dict's keys contains a newline character already!"
            )
        fused_key_bpe_codes = {
            "\n".join(key): float(val) for key, val in orig_bpe.bpe.bpe_codes.items()
        }

        # Initialize a subword Bpe tokenizer
        self.dict = ScriptableDictionaryAgent(
            null_token=orig_dict.null_token,
            end_token=orig_dict.end_token,
            unk_token=orig_dict.unk_token,
            start_token=orig_dict.start_token,
            freq=orig_dict.freq,
            tok2ind=orig_dict.tok2ind,
            ind2tok=orig_dict.ind2tok,
            bpe_add_prefix_space=agent.opt["bpe_add_prefix_space"],
            bpe_encoder={},
            bpe_byte_encoder={},
            fused_key_bpe_ranks={},
            special_tokens=agent._get_special_tokens(),
            subword_bpe_version=orig_bpe.bpe.version,
            fused_bpe_codes=fused_key_bpe_codes,
            subword_bpe_separator=orig_bpe.bpe.separator,
        )

        self.delimiter_tok = agent.history.delimiter_tok
        self.history_size = agent.opt["history_size"]
        if agent.opt.get("history_add_global_end_token", None) is not None:
            self.global_end_token = agent.dict[agent.dict.end_token]
        else:
            self.global_end_token = None
        self.text_truncate = agent.opt.get("text_truncate") or agent.opt["truncate"]
        self.text_truncate = self.text_truncate if self.text_truncate >= 0 else None
        self.class_list = agent.class_list

        agent.model.eval()
        # Create sample inputs for tracing
        sample_tokens = torch.tensor(
            [[1, 2, 3, 4, 5]], dtype=torch.long, device=self.device
        )
        scores = agent.model(sample_tokens)
        _, prediction = torch.max(scores, 1)

        # Do tracing
        self.model = torch.jit.trace(agent.model, sample_tokens)

    def get_device(self):
        return self.model.transformer.embeddings.weight.device

    def parse(self, text: str) -> List[int]:
        return self.dict.txt2vec(text, dict_tokenizer='bpe')

    def forward(self, context: str, max_len: int = 128) -> str:

        history_vecs: List[List[int]] = []
        context_lines = context.split("\n")
        for line in context_lines:
            history_vecs.append(self.parse(line))
        if self.history_size > 0:
            context_lines = context_lines[-self.history_size :]
        # Get full history vec
        text_vecs: List[List[int]] = []
        for vec in history_vecs[:-1]:
            text_vecs += [vec]
            text_vecs += [self.delimiter_tok]
        text_vecs += [history_vecs[-1]]
        if self.global_end_token is not None:
            text_vecs += [[self.global_end_token]]
        # Flatten text_vecs
        flattened_text_vec: List[int] = []
        for vec in text_vecs:
            for token in vec:
                flattened_text_vec.append(token)

        if self.text_truncate is not None:
            truncate_length = self.text_truncate
            if len(flattened_text_vec) > truncate_length:
                flattened_text_vec = flattened_text_vec[-truncate_length:]
        flattened_text_vec = torch.tensor(flattened_text_vec, dtype=torch.long)

        # Added start and end token idx
        flattened_text_vec = torch.cat(
            [
                torch.tensor([self.start_idx], dtype=torch.long).to(self.get_device()),
                flattened_text_vec.to(self.get_device()),
                torch.tensor([self.end_idx], dtype=torch.long).to(self.get_device()),
            ],
            dim=0,
        )
        flattened_text_vec = flattened_text_vec.to(self.get_device())
        batch_text_vec = torch.unsqueeze(flattened_text_vec, dim=0)
        scores = self.model(batch_text_vec)
        _, prediction_id = torch.max(scores, 1)
        preds = self.class_list[prediction_id.squeeze()]

        return preds



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BB3 with R2C2 base.

SDM: Search Decision Model
MDM: Memory Decision Model
SGM: Search Query Generator Model
MGM: Memory Generator Model
MKM: Memory Knowledge Model
CKM: Contextual Knowledge Model
SKM: Search Knowledge Model
MRM: Memory Dialogue Response Model
CRM: Contextual Dialogue Response Model
SRM: Search Dialogue Response Model
"""
from collections import defaultdict
import copy
import torch
from types import MethodType
from typing import List, Tuple, Optional, Dict, Any, Union

from parlai.agents.rag.retrievers import (
    Document,
    RagRetriever,
    BLANK_DOC,
    RetrieverType,
)
from parlai.core.agents import Agent, create_agent, create_agent_from_shared
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.torch_agent import History
import parlai.utils.logging as logging

from projects.seeker.agents.seeker import (
    SeekerAgent,
    ComboFidSearchQueryAgent,
    ComboFidGoldDocumentAgent,
)
from projects.seeker.agents.seeker_modules import (
    combo_fid_retriever_factory,
    ComboFidModel,
    ComboFidSearchQuerySearchEngineRetriever,
)
from projects.seeker.utils import (
    SearchDecision,
    krm_get_batch_context_only_knowledge,
    krm_get_batch_context,
    drm_get_batch_context,
)
from projects.seeker.agents.modular_agent import ModularAgentMixin


import projects.bb3.constants as CONST
from projects.bb3.agents.module import Module
import projects.bb3.tasks.mutators  # noqa: F401
from projects.bb3.agents.utils import clean_text, Decision, MemoryUtils


def bb3_retriever_factory(
    opt: Opt, dictionary: DictionaryAgent, shared=None
) -> Optional[RagRetriever]:
    """
    Build a BB3 retriever, which can handle both memories and search.
    """
    if opt.get('converting'):
        return None
    retriever = RetrieverType(opt['rag_retriever_type'])
    if retriever is RetrieverType.SEARCH_ENGINE:
        return BB3Retriever(opt, dictionary, shared=shared)  # type: ignore
    else:
        return combo_fid_retriever_factory(opt, dictionary, shared)


class BB3Model(ComboFidModel):
    """
    Override the ComboFid model to allow memory computation.
    """

    @classmethod
    def build_retriever(
        cls,
        opt: Opt,
        dictionary: DictionaryAgent,
        retriever_shared: Optional[Dict[str, Any]],
    ) -> Optional[RagRetriever]:
        return bb3_retriever_factory(opt, dictionary, retriever_shared)

    def set_memory(self, memories: List[List[str]]):
        """
        Set retriever's memories.

        :param memories:
            batchsize-length list of memories for each batch item
        """
        assert self.retriever is not None
        self.retriever.set_memory(memories)

    def get_memory(self) -> List[List[str]]:
        """
        Get retriever's memories.

        :return memories:
            return a batchsize-length list of memories for each batch item.
        """
        assert self.retriever is not None
        return self.retriever.get_memory()

    def set_retriever_type(self, r_type: str):
        """
        Informs the retriever whether to retrieve via search queries or long term
        memory.

        :param r_type:
            string indicating whether we're using search or memory.
        """
        assert self.retriever is not None
        self.retriever.retriever_type = r_type


class BB3Retriever(ComboFidSearchQuerySearchEngineRetriever):
    """
    Override the retriever to function as both a long-term memory accessor and internet-
    accessor.
    """

    def __init__(self, opt: Opt, dictionary: DictionaryAgent, shared):
        super().__init__(opt, dictionary, shared=shared)
        self.n_docs = opt['n_docs']
        self.dict = dictionary
        self.memory: List[List[str]] = [[] for _ in range(opt['batchsize'])]
        self._retriever_type = 'search'

    def set_memory(self, memories: List[List[str]]):
        """
        Set retriever memories.
        """
        self.memory = memories

    def get_memory(self) -> List[List[str]]:
        """
        Return retriever memories.
        """
        return self.memory

    @property
    def retriever_type(self) -> str:
        return self._retriever_type

    @retriever_type.setter
    def retriever_type(self, r_type: str):
        self._retriever_type = r_type

    def retrieve_and_score(
        self, query: torch.LongTensor
    ) -> Tuple[List[List[Document]], torch.Tensor]:
        """
        Return the search engine docs if in search mode; otherwise, return memories.
        """
        if self.retriever_type == 'search':
            return super().retrieve_and_score(query)

        top_docs = []
        top_doc_scores = []
        max_n_docs: int = self.n_docs
        for memories in self.memory:
            docs_i = []
            scores_i = []
            for memory in memories:
                docs_i.append(Document(docid='', text=memory, title=''))
                scores_i.append(1)
            # Change this debug later
            max_n_docs = max(max_n_docs, len(docs_i))
            top_docs.append(docs_i)
            top_doc_scores.append(scores_i)
        # Pad with empty docs
        for i in range(len(top_docs)):
            n_empty = max_n_docs - len(top_docs[i])
            if n_empty:
                top_docs[i] = top_docs[i] + [BLANK_DOC] * n_empty
                top_doc_scores[i] = top_doc_scores[i] + [0] * n_empty
        self.top_docs = top_docs
        return top_docs, torch.Tensor(top_doc_scores).to(query.device)


class BB3SubAgentMixin(Agent):
    """
    Override Combo Agent to allow storing memories in the retriever.
    """

    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = 'BlenderBot3'

    def build_model(self) -> BB3Model:
        """
        Build and return BB3Model.
        """
        if self.generation_model == 't5':
            raise RuntimeError('T5 currently not supported')
        else:
            model = BB3Model(self.opt, self.dict)
        return model

    def set_memory(self, memories: List[List[str]]):
        """
        Set retriever's memories.
        """
        self.model_api.set_memory(memories)

    def get_memory(self) -> List[List[str]]:
        """
        Get retriever's memories.
        """
        return self.model_api.get_memory()

    def set_retriever_type(self, r_type: str):
        """
        Set retriever type.
        """
        self.model_api.set_retriever_type(r_type)


class BB3SubSearchAgent(BB3SubAgentMixin, ComboFidSearchQueryAgent):
    pass


class BB3SubGoldAgent(BB3SubAgentMixin, ComboFidGoldDocumentAgent):
    pass


class BlenderBot3Agent(ModularAgentMixin):
    """
    BB3 Agent.

    We DO NOT subclass SeeKeR, and re-implement everything.

    In places, it is noted where things could be abstracted to their own function.
    """

    MODEL_TYPE: str = 'R2C2'

    @classmethod
    def get_additional_agent_args(cls) -> ParlaiParser:
        return SeekerAgent.get_additional_agent_args()

    @classmethod
    def add_additional_subagent_args(cls, parser: ParlaiParser) -> ParlaiParser:
        """
        Setup the subagent args for all submodules.
        """
        additional_agent_parser = cls.get_additional_agent_args()
        for action in additional_agent_parser._actions:
            key = max(action.option_strings, key=lambda x: len(x))
            type = action.type

            for m in Module:
                parser.add_argument(
                    f'--{m.tag()}-{key.strip("-")}', type=type, required=False
                )
        return parser

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Command line args for BB3.
        """
        cls.add_additional_subagent_args(parser)
        group = parser.add_argument_group('SeeKeR Agent Args')
        for module in Module:
            tag = module.tag()
            group.add_argument(
                f'--{tag}-model',
                type=str,
                help=f'agent (not model file) to load for {tag}',
            )
            group.add_argument(
                f'--{tag}-memory-retriever',
                type='bool',
                default=False,
                help='specify for memory retriever',
            )
            if module.r2c2_prompt():
                group.add_argument(
                    f'--{module.message_name()}-control-token',
                    type=str,
                    default=module.r2c2_prompt(),
                    help=f'control token for {tag}',
                )
            if module.is_dialogue():
                group.add_argument(
                    f'--beam-disregard-knowledge-for-{tag}-context-blocking',
                    type='bool',
                    default=False,
                    help=f'If True, disregard the knowledge input for {module.agent_name()} context blocking.',
                )
            elif module.is_knowledge():
                group.add_argument(
                    f'--include-knowledge-in-{tag}-context-blocking',
                    type=bool,
                    default=True,
                    help=f'If True, put the {module.agent_name()} responses in the context for the {module.agent_name()}.',
                )
                group.add_argument(
                    f'--exclude-context-in-{tag}-context-blocking',
                    type=bool,
                    default=False,
                    help=f'Used in conjunction with --include-knowledge-in-{tag}-context-blocking. '
                    f'If specified, only block on the knowledge, and not the concatenation.',
                )
        group.add_argument(
            '--memory-decision-do-access-reply',
            type=str,
            default=CONST.DO_ACCESS_MEMORY,
            help='control token returned by MDM to indicate using memory',
        )
        group.add_argument(
            '--memory-decision-dont-access-reply',
            type=str,
            default=CONST.DONT_ACCESS_MEMORY,
            help='control token returned by MDM to indicate not using memory',
        )
        group.add_argument(
            '--knowledge-conditioning',
            type=str,
            choices=['combined', 'separate', 'both'],
            default='combined',
            help='Specify the way in which the model uses knowledge.\n'
            'combined: condition on all knowledge simultaneously'
            'separate: condition on all knowledge separately, re-ranke later'
            'both: do both combined and separate and re-rank final beam',
        )
        group.add_argument(
            '--combined-dialogue-module',
            type=str,
            choices=[m.tag() for m in Module.dialogue_modules()],
            default=Module.SEARCH_DIALOGUE.tag(),
            help='Specify which agent module handles the final dialogue response',
        )
        parser.add_argument(
            '--ignore-in-session-memories-mkm',
            type='bool',
            default=False,
            help='If true, we do not look at the in-session memories when '
            'generating from the MKM',
        )
        parser.add_argument(
            '--memory-overlap-threshold',
            type=float,
            default=0.0,
            help='Threshold word overlap between memory and incoming text. If '
            'below this threshold, we exclude the memory from consideration.',
        )
        parser.add_argument(
            '--memory-hard-block-for-n-turns',
            type=int,
            default=0,
            help='If set > 0, we block memories from being used for n turns',
        )
        parser.add_argument(
            '--memory-soft-block-decay-factor',
            type=float,
            default=0.0,
            help='If set > 0, we randomly block a memory with probability '
            '(decay ^ n_turns_since_used)',
        )
        # Copied from Seeker
        group.add_argument(
            '--search-decision-do-search-reply',
            type=str,
            default=CONST.DO_SEARCH,
            help='control token returned by SDM to indicate searching',
        )
        group.add_argument(
            '--search-decision-dont-search-reply',
            type=str,
            default=CONST.DO_NOT_SEARCH,
            help='control token returned by SDM to indicate not searching',
        )
        group.add_argument(
            '--all-model-path',
            type=str,
            default=None,
            help='If specified, load all models with this path',
        )
        group.add_argument(
            '--search-decision',
            type=str,
            default=Decision.COMPUTE.value,
            choices=[s.value for s in SearchDecision],
        )
        group.add_argument(
            '--memory-decision',
            type=str,
            default=Decision.COMPUTE.value,
            choices=[s.value for s in SearchDecision],
        )
        group.add_argument(
            '--memory-decision-use-memories',
            type='bool',
            default=True,
            help='If true, the memory decision module will have access to memories when making the decision',
        )
        group.add_argument(
            '--contextual-knowledge-decision',
            type=str,
            default=Decision.COMPUTE.value,
            choices=[s.value for s in SearchDecision],
        )
        group.add_argument(
            '--inject-query-string',
            type=str,
            default=None,
            help='If set, this string is appended to all search queries.',
        )
        group.add_argument(
            '--search-server', type=str, default=None, help='search server to use.'
        )
        group.add_argument(
            '--serializable-output',
            type='bool',
            default=False,
            help='Whether to make output serializable. Specify True when using e.g. mechanical turk.',
        )

        return parser

    def __init__(self, opt, shared=None):
        self.id = 'BlenderBot3'
        opt = copy.deepcopy(opt)
        self.opt = opt
        one_model = opt.get('all_model_path') or opt.get('model_file')
        assert one_model, "Must specify a model file for this agent."
        for k in Module:
            opt[k.model_file_path_key()] = one_model
        self._construct_subagent_opts(opt)

        self.agents: Dict[Union[str, Module], Agent] = {}
        self.clones: Dict[Union[str, Module], List[Agent]] = {}
        if not shared:
            agent_opts = self.opts[Module.SEARCH_KNOWLEDGE]
            agent = self._init_top_agent(agent_opts)
            # two mappings; one for code access, one for debug access.
            self.agents[Module.SEARCH_KNOWLEDGE.agent_name()] = agent
            self.agents[Module.SEARCH_KNOWLEDGE] = agent
            logging.verbose(f"options for {Module.SEARCH_KNOWLEDGE.agent_name()}")
            if logging.logger.isEnabledFor(logging.VERBOSE):
                Opt(self.agents[Module.SEARCH_KNOWLEDGE].opt).log()
        else:
            agent = create_agent_from_shared(shared['search_knowledge_agent_share'])
            self.agents[Module.SEARCH_KNOWLEDGE.agent_name()] = agent
            self.agents[Module.SEARCH_KNOWLEDGE] = agent

        if hasattr(agent, 'dictionary'):
            # OPT
            self.dictionary = agent.dictionary
        else:
            # R2C2
            self.dictionary = agent.dict
        # Memories is a mapping from memory to turns_since_used.
        self.memories: Dict[str, int] = {}
        self.in_session_memories = set()
        self.search_knowledge_responses = ['__SILENCE__']
        self.memory_knowledge_responses = ['__SILENCE__']
        self.contextual_knowledge_responses = ['__SILENCE__']
        self.search_decision = Decision(opt['search_decision'])
        self.memory_decision = Decision(opt['memory_decision'])
        self.contextual_knowledge_decision = Decision(
            opt['contextual_knowledge_decision']
        )
        self.inject_query_string = opt.get('inject_query_string', '')
        self.knowledge_conditioning = opt['knowledge_conditioning']
        self.combined_dialogue_module = Module(opt['combined_dialogue_module'])
        self.memory_utils = MemoryUtils

        for m in Module:
            agent = self._init_shared_model(m)
            self.agents[m.agent_name()], self.agents[m] = agent, agent
            if m.is_dialogue():
                self.clones[m.agent_name()], self.clones[m] = [agent], [agent]

        if not shared:
            # the main agent needs batchsize-many clones for handling history during batch act.
            for m in Module.dialogue_modules():
                self.clones[m.agent_name()] += [
                    self.agents[m.agent_name()].clone()
                    for _ in range(opt.get('batchsize', 1) - 1)
                ]
                self.clones[m] = self.clones[m.agent_name()]

        self._apply_context_blocking_patches()

        for m in Module:
            if m is Module.MEMORY_DECISION and opt['memory_decision_use_memories']:
                continue
            if m.is_one_turn_history():
                try:
                    assert (
                        self.agents[m].history.size == 1
                    ), f"wrong history size! set --{m.tag()}-history-size 1"
                except AttributeError:
                    pass

        super().__init__(opt, shared)

    def _init_top_agent(self, opt: Opt) -> Agent:
        """
        Initialize the toplevel agent.
        """
        return create_agent(opt, requireModelExists=True)

    def _apply_context_blocking_patches(self):
        """
        Optionally monkey-patch custom context blocking functions.
        """
        for m in Module:
            agent = self.agents[m]
            if m.is_knowledge():
                if self.opt[f'include_knowledge_in_{m.tag()}_context_blocking']:
                    orig_fun = agent._get_batch_context
                    if self.opt[f'exclude_context_in_{m.tag()}_context_blocking']:
                        agent._get_batch_context = MethodType(
                            lambda self, batch: krm_get_batch_context_only_knowledge(
                                self, batch, orig_fun
                            ),
                            agent,
                        )
                    else:
                        agent._get_batch_context = MethodType(
                            lambda self, batch: krm_get_batch_context(
                                self, batch, orig_fun
                            ),
                            agent,
                        )
            elif m.is_dialogue():
                if self.opt[f'beam_disregard_knowledge_for_{m.tag()}_context_blocking']:
                    orig_fun = agent._get_batch_context
                    agent._get_batch_context = MethodType(
                        lambda self, batch: drm_get_batch_context(
                            self, batch, orig_fun=orig_fun
                        ),
                        agent,
                    )

    @property
    def history(self) -> History:
        return self.agents[Module.SEARCH_DIALOGUE].history

    @property
    def knowledge_agent_history(self) -> History:
        return self.agents[Module.SEARCH_KNOWLEDGE].history

    def get_history_str(self) -> str:
        return self.history.get_history_str()

    def reset(self, clones_only: bool = False):
        """
        Override to reset all sub agents.

        :param clones_only:
            if true, only reset the dialogue clones
        """
        for agents in self.clones.values():
            for a in agents:
                a.reset()
        if not clones_only:
            for agent in self.agents.values():
                agent.reset()
            self.search_knowledge_responses = ['__SILENCE__']
            self.contextual_knowledge_responses = ['__SILENCE__']
            self.memory_knowledge_responses = ['__SILENCE__']
            self.memories = {}
            self.in_session_memories = set()

    def _construct_subagent_opts(self, opt: Opt):
        """
        Construct opts for each sub agent.

        :param opt:
            original Opt.
        """
        if opt['search_server']:
            opt['skm_search_server'] = opt['search_server']
            if 'override' in opt:
                opt['override'][f'{Module.SEARCH_KNOWLEDGE.tag()}_search_server'] = opt[
                    'search_server'
                ]
        self.opts = {}
        self.opts['init'] = opt
        override_opts = defaultdict(dict)
        for k, v in opt['override'].items():
            k_set = False
            for m in Module:
                if k.startswith(f'{m.tag()}_'):
                    override_opts[m][k] = v
                    k_set = True
                    break
            if not k_set:
                override_opts['general'][k] = v
        self.opts['override'] = override_opts
        for m in Module:
            if f'{m.tag()}_interactive_mode' not in override_opts[m]:
                override_opts[m][f'{m.tag()}_interactive_mode'] = opt.get(
                    'interactive_mode', False
                )

        for m in Module:
            filename = opt[m.model_file_path_key()]
            if not filename:
                continue
            self.opts[m] = self._get_subagent_opt(
                filename=filename,
                specific_override_args=override_opts[m],
                general_override_args=override_opts['general'],
            )
            self.opts[m]['model_file'] = filename
            self.opts[m]['override']['model_file'] = filename

    def _init_shared_model(self, opt_key: Module) -> Agent:
        """
        Initialize shared version of a model, for sub modules.
        """
        return super().init_shared_model(
            self.opts[opt_key], self.agents[Module.SEARCH_KNOWLEDGE]
        )

    def _get_subagent_opt(
        self,
        filename: str,
        specific_override_args: Dict[str, Any],
        general_override_args: Dict[str, Any],
    ) -> Opt:
        """
        Return the specific subagent opt parameters.
        """
        return super().get_subagent_opt(
            self.opt['datapath'],
            filename,
            specific_override_args,
            general_override_args,
        )

    def share(self):
        """
        Share the top agent.
        """
        shared = super().share()
        shared['search_knowledge_agent_share'] = self.agents[
            Module.SEARCH_KNOWLEDGE
        ].share()
        return shared

    def get_mdm_observation(self, ag_obs: Message) -> Message:
        """
        Add memories to the memory decision observation.

        :param ag_obs:
            incoming mdm observation

        :return mdm_obs:
            return mdm observation with memories in the context.
        """
        # Reset the history for one-turn agents
        # for now, protect this heavily, as this is being added for mdm only
        self.agents[Module.MEMORY_DECISION].reset()
        if self.memories:
            self_memories = [
                m.replace('your persona: ', '')
                for m in self.memories
                if m.startswith('your')
            ]
            partner_memories = [
                m.replace("partner's persona: ", '')
                for m in self.memories
                if m.startswith('partner')
            ]
            memories = f"your persona: {' '.join(self_memories)}\npartner's persona: {' '.join(partner_memories)}"
            ag_obs.force_set('text', '\n'.join([memories, ag_obs['text']]))
        return ag_obs

    def _get_memory_heuristic_values(self) -> Dict[str, Union[str, float, bool]]:
        """
        Extract heuristics from self.opt.
        """
        return {
            'ignore_in_session_memories': self.opt.get(
                'ignore_in_session_memories_mkm', False
            ),
            'memory_overlap_threshold': self.opt.get('memory_overlap_threshold', 0.0),
            'memory_hard_block_for_n_turns': self.opt.get(
                'memory_hard_block_for_n_turns', 0
            ),
            'memory_soft_block_decay_factor': self.opt.get(
                'memory_soft_block_decay_factor', 0.0
            ),
        }

    def get_available_memories(
        self, observations: List[Dict[Union[str, Module], Message]]
    ) -> List[List[str]]:
        """
        Get available memories for the agent.

        :param observations:
            incoming batch observations

        :return memories:
            return a list of memories for each batch item.
        """
        return [
            self.memory_utils.get_available_memories(
                o['raw']['text'],
                o['raw']['memories'],
                o['raw']['in_session_memories'],
                self.dictionary,
                **self._get_memory_heuristic_values(),
            )
            for o in observations
        ]

    def observe(self, observation: Message) -> Dict[Module, Message]:
        """
        Observe in all modules besides dialogue ones, which require outputs from the
        knowledge stage.

        :param observation:
            incoming message

        :return self.observation:
            returned observation is actually a dictionary mapping
            module name to the corresponding observation
        """
        observations = {}
        if not isinstance(observation, Message):
            observation = Message(observation)
        for key in ['label_candidates', 'knowledge']:
            # Delete unnecessarily large keys
            observation.pop(key, '')

        raw_observation = copy.deepcopy(observation)
        raw_observation['memories'] = self.memories
        raw_observation['in_session_memories'] = self.in_session_memories
        observations['raw'] = raw_observation

        if observation.get('episode_done'):
            self.search_knowledge_responses = ['__SILENCE__']
            self.contextual_knowledge_responses = ['__SILENCE__']
            self.memory_knowledge_responses = ['__SILENCE__']

        for m in Module:
            # observe for all non-dialogue agents
            if m.is_dialogue():
                continue
            ag_obs = copy.deepcopy(observation)
            if self.opt[f"{m.message_name()}_control_token"]:
                ag_obs.force_set(
                    'temp_history', f" {self.opt[f'{m.message_name()}_control_token']}"
                )
            if m.skip_search():
                ag_obs.force_set('skip_retrieval', True)
            if m is Module.CONTEXTUAL_KNOWLEDGE:
                ag_obs['prior_knowledge_responses'] = ' '.join(
                    self.contextual_knowledge_responses
                )
            elif m is Module.SEARCH_KNOWLEDGE:
                ag_obs['prior_knowledge_responses'] = ' '.join(
                    self.search_knowledge_responses
                )
            elif m is Module.MEMORY_KNOWLEDGE:
                ag_obs['prior_knowledge_responses'] = ' '.join(
                    self.memory_knowledge_responses
                )
            if m is Module.MEMORY_DECISION and self.opt['memory_decision_use_memories']:
                ag_obs = self.get_mdm_observation(ag_obs)

            observations[m] = self.agents[m].observe(ag_obs)

        self.observations = observations
        return observations

    def batch_act_decision(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        module: Module,
        agent: Agent,
    ) -> Tuple[List[Message], List[int]]:
        """
        Decision agent batch act.

        :param observations:
            observations for batch act.
        :param module:
            module making the decision

        :return (batch_reply, indices):
            batch_reply: reply from the decision agent
            indices: batch indices with which to use search/memory.
        """
        indices = []
        batch_reply = [{} for _ in range(len(observations))]
        if module is Module.SEARCH_DECISION:
            do = self.opt['search_decision_do_search_reply']
            dont = self.opt['search_decision_dont_search_reply']
            decision = self.search_decision
        else:
            assert module is Module.MEMORY_DECISION
            do = self.opt['memory_decision_do_access_reply']
            dont = self.opt['memory_decision_dont_access_reply']
            decision = self.memory_decision

        if decision is Decision.ALWAYS:
            indices = list(range(len(observations)))
        elif decision is Decision.NEVER:
            indices = []
        else:
            assert decision is Decision.COMPUTE
            assert agent
            batch_reply = agent.batch_act([o[module] for o in observations])
            for i, reply in enumerate(batch_reply):
                logging.debug(f"Example {i}, {module.agent_name()}: {reply['text']}")
                if reply['text'] == do or reply['text'].startswith(do):
                    indices.append(i)
                elif reply['text'] == dont or reply['text'].startswith(do):
                    continue
                else:
                    logging.error(
                        f"Decision Reply: {reply['text']}; defaulting to no search/memory"
                    )

        return batch_reply, indices

    def batch_act_sgm(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        search_indices: List[int],
        agent: Agent,
    ) -> List[Message]:
        """
        Search Query Generator batch act.

        :param observations:
            list of observations
        :param search_indices:
            list of batch indices for which search is required.
        :param agent:
            search query generator agent

        :return batch_reply:
            return the batch reply from the search query agent
        """
        return super().batch_act_search_query_generation(
            observations,
            [o[Module.SEARCH_QUERY] for o in observations],
            search_indices,
            agent,
            self.agents[Module.SEARCH_KNOWLEDGE],
            self.inject_query_string,
        )

    def batch_act_mgm(
        self,
        observations: Optional[List[Dict[Union[str, Module], Message]]] = None,
        self_messages: Optional[List[Message]] = None,
        agent: Optional[Agent] = None,
    ) -> List[Message]:
        """
        Memory Generator batch act.

        :param observations:
            list of observations, where the text is the partner's message
        :param self_messages:
            list of messages that the bot generated.
        :param agent:
            memory generator agent

        :return batch_reply:
            return the batch reply from the search query agent
        """
        assert agent is not None
        if observations is not None:
            batch_reply_mgm = agent.batch_act(
                [o[Module.MEMORY_GENERATOR] for o in observations]
            )
            logging.debug(f"Partner Memories: {[a['text'] for a in batch_reply_mgm]}")
        else:
            assert self_messages is not None
            control_token = self.opt[
                f"{Module.MEMORY_GENERATOR.message_name()}_control_token"
            ]
            batch_reply_mgm = agent.batch_respond(
                [
                    Message(
                        {
                            'text': f"{m['text']} {control_token}"
                            if control_token
                            else m['text'],
                            'episode_done': True,
                        }
                    )
                    for m in self_messages
                ]
            )
            batch_reply_mgm = [Message({'text': t}) for t in batch_reply_mgm]
            logging.debug(f"Self Memories: {[a['text'] for a in batch_reply_mgm]}")
        agent.reset()
        return batch_reply_mgm

    def batch_act_knowledge(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        search_indices: List[int],
        memory_indices: List[int],
        contextual_indices: List[int],
        batch_agents: Dict[Module, Agent],
        top_docs: Optional[List[List[Document]]] = None,
        top_memories: Optional[List[List[str]]] = None,
    ) -> List[Message]:
        """
        Batch act with Knowledge Models.

        :param observations:
            list of observations
        :param search_indices:
            list of indices for which we search.
        :param memory_indices:
            list of indices for which we access memory.
        :param contextual_indices:
            list of indices for which we access knowledge from context.
        :param top_docs:
            list of top documents; send if not normally returned in the batch act
        :param top_memories:
            list of top memories; send if not normally returned in the batch act

        :return batch_reply:
            batch_reply: batch reply from knowledge modules
        """
        batch_reply_knowledge = [Message({}) for _ in range(len(observations))]
        ckm_obs, skm_obs, mkm_obs = [], [], []
        for i, o in enumerate(observations):
            if i in contextual_indices:
                ckm_obs.append(o[Module.CONTEXTUAL_KNOWLEDGE])
            if i in search_indices:
                skm_obs.append(o[Module.SEARCH_KNOWLEDGE])
            if i in memory_indices:
                mkm_obs.append(o[Module.MEMORY_KNOWLEDGE])

        batch_reply_ckm = batch_agents[Module.CONTEXTUAL_KNOWLEDGE].batch_act(ckm_obs)
        batch_agents[Module.SEARCH_KNOWLEDGE].set_retriever_type('search')
        batch_reply_skm = batch_agents[Module.SEARCH_KNOWLEDGE].batch_act(skm_obs)
        batch_agents[Module.MEMORY_KNOWLEDGE].set_retriever_type('memory')
        batch_reply_mkm = batch_agents[Module.MEMORY_KNOWLEDGE].batch_act(mkm_obs)

        search_offset = 0
        memory_offset = 0
        contextual_offset = 0
        for i, message in enumerate(batch_reply_knowledge):
            if i in contextual_indices:
                message[Module.CONTEXTUAL_KNOWLEDGE.message_name()] = batch_reply_ckm[
                    contextual_offset
                ]['text']
                logging.debug(
                    f"Contextual KNOWLEDGE for example {i}: {message[Module.CONTEXTUAL_KNOWLEDGE.message_name()]}"
                )
                contextual_offset += 1
            if i in search_indices:
                message[Module.SEARCH_KNOWLEDGE.message_name()] = batch_reply_skm[
                    search_offset
                ]['text']
                logging.debug(
                    f"Search KNOWLEDGE for example {i}: {message[Module.SEARCH_KNOWLEDGE.message_name()]}"
                )
                docs = (
                    top_docs[search_offset]
                    if top_docs is not None
                    else batch_reply_skm[search_offset]['top_docs']
                )
                message[f'{Module.SEARCH_KNOWLEDGE.message_name()}_top_docs'] = docs
                search_offset += 1
            if i in memory_indices:
                message[Module.MEMORY_KNOWLEDGE.message_name()] = batch_reply_mkm[
                    memory_offset
                ]['text']
                logging.debug(
                    f"Memory KNOWLEDGE for example {i}: {message[Module.MEMORY_KNOWLEDGE.message_name()]}"
                )
                docs = (
                    top_memories[memory_offset]
                    if top_memories is not None
                    else batch_reply_mkm[memory_offset]['top_docs']
                )
                message[f'{Module.MEMORY_KNOWLEDGE.message_name()}_top_docs'] = docs
                memory_offset += 1

        return batch_reply_knowledge

    def batch_act_dialogue_combined(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        batch_reply_knowledge: List[Message],
    ) -> List[Message]:
        """
        Dialogue batch act.

        Combine all available knowledge sources for generation.

        :param observations:
            observations from self.observe
        :param batch_reply_knowledge:
            batch reply from the knowledge module

        :return batch_reply_dialogue:
            return the reply from the dialogue models.
        """
        full_text = [
            clean_text(
                o[Module.CONTEXTUAL_KNOWLEDGE].get('full_text', o.get('text', ''))
            )
            for o in observations
        ]
        clones = self.clones[Module.SEARCH_DIALOGUE]
        agent = self.agents[self.combined_dialogue_module]
        dialogue_agent_observations = []
        for i, (obs, knowledge_obs) in enumerate(
            zip(observations, batch_reply_knowledge)
        ):
            temp_history = '\n'
            srm_obs = copy.deepcopy(obs['raw'])
            for m in Module.knowledge_modules():
                if knowledge_obs.get(m.message_name()):
                    tokens = m.special_tokens()
                    temp_history += (
                        f"{tokens[0]} {knowledge_obs[m.message_name()]} {tokens[1]}"
                    )
            assert temp_history
            srm_obs.force_set('temp_history', temp_history)
            srm_obs.force_set('skip_retrieval', True)
            if not clones[i].history.get_history_str():
                srm_obs.force_set('text', full_text[i])
            dialogue_agent_observations.append(clones[i].observe(srm_obs))
        batch_reply_srm = agent.batch_act(dialogue_agent_observations)
        for i, obs in enumerate(batch_reply_srm):
            logging.debug(
                f"Combined DIALOGUE response for example {i}: {obs['text']}; score: {obs['beam_texts'][0][-1]:.2f}"
            )
            clones[i].self_observe(obs)
            # manually clear
            clones[i].history.temp_history = None

        return batch_reply_srm

    def batch_act_dialogue_separate(
        self,
        observations: List[Dict[Union[Module, str], Message]],
        batch_reply_knowledge: List[Message],
        search_indices: List[int],
        memory_indices: List[int],
        contextual_indices: List[int],
    ) -> List[Message]:
        """
        Dialogue batch act.

        Separately generate dialogue responses for each module with knowledge.

        Then, re-rank according to beam likelihood.

        :param observations:
            observations from self.observe
        :param batch_reply_knowledge:
            batch reply from the knowledge module
        :param search_indices:
            indices that used search
        :param memory_indices:
            indices that used memory
        :param contextual_indices:
            indices that used contextual knowledge

        :return batch_reply_dialogue:
            return the reply from the dialogue module.
        """
        full_text = [
            clean_text(
                o[Module.CONTEXTUAL_KNOWLEDGE].get('full_text', o.get('text', ''))
            )
            for o in observations
        ]
        srm = self.agents[Module.SEARCH_DIALOGUE]
        mrm = self.agents[Module.MEMORY_DIALOGUE]
        crm = self.agents[Module.CONTEXTUAL_DIALOGUE]
        srm_clones = self.clones[Module.SEARCH_DIALOGUE]
        crm_clones = self.clones[Module.MEMORY_DIALOGUE]
        mrm_clones = self.clones[Module.CONTEXTUAL_DIALOGUE]

        contextual_observations, search_observations, memory_observations = [], [], []
        batch_reply_crm, batch_reply_srm, batch_reply_mrm = [], [], []

        # First, we generate the observations given the knowledge outputs.
        for i, (obs, knowledge_obs) in enumerate(
            zip(observations, batch_reply_knowledge)
        ):
            dialogue_obs = copy.deepcopy(obs['raw'])
            dialogue_obs.force_set('skip_retrieval', True)
            if i in search_indices:
                factual_obs = copy.deepcopy(dialogue_obs)
                tokens = Module.SEARCH_KNOWLEDGE.special_tokens()
                factual_obs.force_set(
                    'temp_history',
                    f"\n{tokens[0]} {knowledge_obs[Module.SEARCH_KNOWLEDGE.message_name()]} {tokens[1]}",
                )
                if not srm_clones[i].history.get_history_str():
                    factual_obs.force_set('text', full_text[i])
                search_observations.append(srm_clones[i].observe(factual_obs))
            if i in memory_indices:
                memory_obs = copy.deepcopy(dialogue_obs)
                tokens = Module.MEMORY_KNOWLEDGE.special_tokens()
                memory_obs.force_set(
                    'temp_history',
                    f"\n{tokens[0]} {knowledge_obs[Module.MEMORY_KNOWLEDGE.message_name()]} {tokens[1]}",
                )
                if not mrm_clones[i].history.get_history_str():
                    memory_obs.force_set('text', full_text[i])
                memory_observations.append(mrm_clones[i].observe(memory_obs))
            if i in contextual_indices:
                contextual_obs = copy.deepcopy(dialogue_obs)
                tokens = Module.CONTEXTUAL_KNOWLEDGE.special_tokens()
                contextual_obs.force_set(
                    'temp_history',
                    f"\n{tokens[0]} {knowledge_obs[Module.CONTEXTUAL_KNOWLEDGE.message_name()]} {tokens[1]}",
                )
                if not crm_clones[i].history.get_history_str():
                    contextual_obs.force_set('text', full_text[i])
                contextual_observations.append(crm_clones[i].observe(contextual_obs))

        if search_observations:
            batch_reply_srm = srm.batch_act(search_observations)
        if memory_observations:
            batch_reply_mrm = mrm.batch_act(memory_observations)
        if contextual_observations:
            batch_reply_crm = crm.batch_act(contextual_observations)

        batch_reply_dialogue = [Message({}) for _ in range(len(observations))]
        search_offset = 0
        memory_offset = 0
        contextual_offset = 0
        # Second, we generate the dialogue responses.
        for i in range(len(observations)):
            if i in search_indices:
                search_act = batch_reply_srm[search_offset]
                srm_clones[search_offset].self_observe(search_act)
                batch_reply_dialogue[i][
                    Module.SEARCH_DIALOGUE.message_name()
                ] = search_act['text']
                batch_reply_dialogue[i][
                    f"{Module.SEARCH_DIALOGUE.message_name()}_score"
                ] = search_act['beam_texts'][0][-1]
                search_offset += 1
                logging.debug(
                    f"Search DIALOGUE response for {i}: {search_act['text']}; score: {search_act['beam_texts'][0][-1]:.2f}"
                )
            if i in memory_indices:
                memory_act = batch_reply_mrm[memory_offset]
                mrm_clones[memory_offset].self_observe(memory_act)
                batch_reply_dialogue[i][
                    Module.MEMORY_DIALOGUE.message_name()
                ] = memory_act['text']
                batch_reply_dialogue[i][
                    f"{Module.MEMORY_DIALOGUE.message_name()}_score"
                ] = memory_act['beam_texts'][0][-1]
                memory_offset += 1
                logging.debug(
                    f"Memory DIALOGUE response for {i}: {memory_act['text']}; score: {memory_act['beam_texts'][0][-1]:.2f}"
                )
            if i in contextual_indices:
                contextual_act = batch_reply_crm[contextual_offset]
                crm_clones[contextual_offset].self_observe(contextual_act)
                batch_reply_dialogue[i][
                    Module.CONTEXTUAL_DIALOGUE.message_name()
                ] = contextual_act['text']
                batch_reply_dialogue[i][
                    f"{Module.CONTEXTUAL_DIALOGUE.message_name()}_score"
                ] = contextual_act['beam_texts'][0][-1]
                contextual_offset += 1
                logging.debug(
                    f"Contextual DIALOGUE response for {i}: {contextual_act['text']}; score: {contextual_act['beam_texts'][0][-1]:.2f}"
                )
            # manually clear
            srm_clones[i].history.temp_history = None
            mrm_clones[i].history.temp_history = None
            crm_clones[i].history.temp_history = None

        # Third, we re-rank according to beam likelihood.
        for reply in batch_reply_dialogue:
            options, scores = [], []
            for i, m in enumerate(
                [
                    Module.SEARCH_DIALOGUE,
                    Module.MEMORY_DIALOGUE,
                    Module.CONTEXTUAL_DIALOGUE,
                ]
            ):
                options.append(reply.get(m.message_name(), ''))
                scores.append(
                    (i, reply.get(f"{m.message_name()}_score", -float('inf')))
                )
            max_score = max(scores, key=lambda x: x[-1])
            reply['text'] = options[max_score[0]]
            reply['max_score'] = max_score[1]

        return batch_reply_dialogue

    def collate_batch_acts(
        self,
        batch_reply_sdm: List[Message],
        batch_reply_mdm: List[Message],
        batch_reply_sgm: List[Message],
        batch_reply_mgm_self: List[Message],
        batch_reply_mgm_partner: List[Message],
        batch_reply_knowledge: List[Message],
        batch_reply_dialogue: List[Message],
        available_memory: Union[List[List[str]], List[Dict[str, int]]],
    ) -> List[Message]:
        """
        Collate all of the batch acts from the various modules.

        :param batch_reply_X:
            batch reply from module X
        :param available_memory:
            list of memories for each batch item

        :return batch_reply:
            return the agent's final batch reply
        """
        final_batch_reply = []
        for sdm, mdm, sgm, mgm_self, mgm_partner, km, srm, mems in zip(
            batch_reply_sdm,
            batch_reply_mdm,
            batch_reply_sgm,
            batch_reply_mgm_self,
            batch_reply_mgm_partner,
            batch_reply_knowledge,
            batch_reply_dialogue,
            available_memory,
        ):
            if srm.is_padding():
                continue
            reply = Message(
                {
                    k: v
                    for k, v in srm.items()
                    if k not in ['top_docs']  # leave as list for future use cases
                }
            )
            reply.force_set(Module.SEARCH_DECISION.message_name(), sdm.get('text', ''))
            reply.force_set(Module.MEMORY_DECISION.message_name(), mdm.get('text', ''))
            reply.force_set(Module.SEARCH_QUERY.message_name(), sgm.get('text', ''))
            reply.force_set(
                f'{Module.MEMORY_GENERATOR.message_name()}_self',
                mgm_self.get('text', ''),
            )
            reply.force_set(
                f'{Module.MEMORY_GENERATOR.message_name()}_partner',
                mgm_partner.get('text', ''),
            )
            mems = copy.deepcopy(mems)
            for message, person in zip([mgm_self, mgm_partner], ['self', 'partner']):
                if self.memory_utils.is_valid_memory(
                    mems,
                    message.get('text', ''),
                    self.memory_utils.get_memory_prefix(person, self.MODEL_TYPE),
                ):
                    mems = self.memory_utils.add_memory(
                        self.memory_utils.add_memory_prefix(
                            message['text'], person, self.MODEL_TYPE
                        ),
                        mems,
                    )
            reply.force_set('memories', mems)
            reply.force_set(
                Module.SEARCH_KNOWLEDGE.message_name(),
                km.get(Module.SEARCH_KNOWLEDGE.message_name(), ''),
            )
            reply.force_set(
                Module.CONTEXTUAL_KNOWLEDGE.message_name(),
                km.get(Module.CONTEXTUAL_KNOWLEDGE.message_name(), ''),
            )
            reply.force_set(
                Module.MEMORY_KNOWLEDGE.message_name(),
                km.get(Module.MEMORY_KNOWLEDGE.message_name(), ''),
            )
            for m in Module:
                # set all the knowledge responses
                if m.is_knowledge():
                    reply.force_set(m.message_name(), km.get(m.message_name(), ''))
                # if separate, set all of the dialogue responses as well
                elif m.is_dialogue():
                    reply.force_set(m.message_name(), srm.get(m.message_name(), ''))
                    reply.force_set(
                        f"{m.message_name()}_score",
                        reply.get(f"{m.message_name()}_score", -float('inf')),
                    )

                if not m.skip_search():
                    docs = km.get(
                        f'{m.message_name()}_top_docs', [Document("", "", "")]
                    )
                    reply.force_set(
                        f'{m.message_name()}_doc_titles', [d.get_title() for d in docs]
                    )
                    reply.force_set(
                        f'{m.message_name()}_doc_content', [d.get_text() for d in docs]
                    )
                    reply.force_set(
                        f'{m.message_name()}_doc_urls', [d.get_id() for d in docs]
                    )
            logging.debug(reply)
            final_batch_reply.append(reply)

        if self.opt['serializable_output']:
            for i in range(len(final_batch_reply)):
                final_batch_reply[i] = Message(
                    {
                        k: v
                        for k, v in final_batch_reply[i].items()
                        if k in ['text', 'id', 'episode_done']
                    }
                )
        return final_batch_reply

    def batch_act(
        self, observations: List[Dict[Union[str, Module], Message]]
    ) -> List[Message]:
        """
        Full batch_act pipeline.

        :param observations:
            batchsize-length list of observations from self.observe

        :return reply:
            return batchsize-length list of final replies.
        """
        # First, determine whether we're searching or accessing memory
        all_memory: List[Dict[str, int]] = [o['raw']['memories'] for o in observations]
        try:
            memory_to_set = self.get_available_memories(observations)
            self.agents[Module.MEMORY_KNOWLEDGE].set_memory(memory_to_set)
            available_memory = self.agents[Module.MEMORY_KNOWLEDGE].get_memory()
        except AttributeError:
            # Gold Docs
            available_memory = [[]] * len(observations)
            pass
        batch_reply_sdm, search_indices = self.batch_act_decision(
            observations, Module.SEARCH_DECISION, self.agents[Module.SEARCH_DECISION]
        )
        batch_reply_mdm, memory_indices = self.batch_act_decision(
            observations, Module.MEMORY_DECISION, self.agents[Module.MEMORY_DECISION]
        )
        memory_indices = [i for i in memory_indices if available_memory[i]]
        if self.contextual_knowledge_decision is Decision.ALWAYS:
            contextual_indices = list(range(len(observations)))
        elif self.contextual_knowledge_decision is Decision.NEVER:
            contextual_indices = []
        else:
            assert self.contextual_knowledge_decision is Decision.COMPUTE
            contextual_indices = [
                i
                for i in list(range(len(observations)))
                if i not in memory_indices + search_indices
            ]

        # Second, generate search queries and new memories
        batch_reply_sgm = self.batch_act_sgm(
            observations, search_indices, self.agents[Module.SEARCH_QUERY]
        )
        batch_reply_mgm_partner = self.batch_act_mgm(
            observations=observations, agent=self.agents[Module.MEMORY_GENERATOR]
        )

        # Third, generate the knowledge sentences
        batch_reply_knowledge = self.batch_act_knowledge(
            observations,
            search_indices,
            memory_indices,
            contextual_indices,
            {m: self.agents[m] for m in Module if m.is_knowledge()},
        )

        # Fourth, generate the dialogue response!
        if self.knowledge_conditioning == 'combined':
            batch_reply_dialogue = self.batch_act_dialogue_combined(
                observations, batch_reply_knowledge
            )
        elif self.knowledge_conditioning == 'separate':
            batch_reply_dialogue = self.batch_act_dialogue_separate(
                observations,
                batch_reply_knowledge,
                search_indices,
                memory_indices,
                contextual_indices,
            )
        else:
            assert self.knowledge_conditioning == 'both'
            reply_combined = self.batch_act_dialogue_combined(
                observations, batch_reply_knowledge
            )
            self.reset(clones_only=True)
            reply_separate = self.batch_act_dialogue_separate(
                observations,
                batch_reply_knowledge,
                search_indices,
                memory_indices,
                contextual_indices,
            )
            batch_reply_dialogue = []
            for r_c, r_s in zip(reply_combined, reply_separate):
                reply = r_c
                reply_score = reply['beam_texts'][0][-1]
                max_seperate_score = r_s['max_score']
                if max_seperate_score > reply_score:
                    reply.force_set('text', r_s['text'])
                batch_reply_dialogue.append(reply)

        # Fifth, generate new memories
        batch_reply_mgm_self = self.batch_act_mgm(
            self_messages=batch_reply_dialogue,
            agent=self.agents[Module.MEMORY_GENERATOR],
        )

        # Sixth, combine them all in the srm batch reply.
        final_batch_reply = self.collate_batch_acts(
            batch_reply_sdm,
            batch_reply_mdm,
            batch_reply_sgm,
            batch_reply_mgm_self,
            batch_reply_mgm_partner,
            batch_reply_knowledge,
            batch_reply_dialogue,
            all_memory,
        )

        return final_batch_reply

    def act(self):
        """
        Call batch_act with the singleton batch.
        """
        response = self.batch_act([self.observations])[0]
        self.self_observe(response)
        return response

    def self_observe(self, self_message: Message):
        """
        Override TA.self_observe.

        Make sure all agents have same history.
        """
        self.agents[Module.SEARCH_KNOWLEDGE].self_observe(self_message)
        self.search_knowledge_responses.append(
            self_message.get(Module.SEARCH_KNOWLEDGE.message_name(), '')
        )
        self.contextual_knowledge_responses.append(
            self_message.get(Module.CONTEXTUAL_KNOWLEDGE.message_name(), '')
        )
        self.memory_knowledge_responses.append(
            self_message.get(Module.MEMORY_KNOWLEDGE.message_name(), '')
        )
        self.self_observe_memory(self_message)
        observation = {
            'text': clean_text(
                self.agents[Module.SEARCH_KNOWLEDGE].history.get_history_str() or ''
            )
        }
        for agent_name, agent in self.agents.items():
            if isinstance(agent, list):
                for a in agent:
                    a.reset()
                    a.history.update_history(
                        observation, temp_history=a.get_temp_history(observation)
                    )
            elif agent_name != Module.SEARCH_KNOWLEDGE.message_name():
                agent.reset()
                agent.history.update_history(
                    observation, temp_history=agent.get_temp_history(observation)
                )

    def self_observe_memory(self, self_message: Message):
        """
        Self observe memory for person.

        First, add new memories to the memory.
        Second, update memory usage.

        :param memory:
            potential memory
        :param person:
            whose memory it is
        """
        # add new memories
        memory_key = Module.MEMORY_GENERATOR.message_name()
        for person in ['self', 'partner']:
            memory_candidate = self_message.get(f"{memory_key}_{person}")
            if not memory_candidate:
                continue
            if self.memory_utils.is_valid_memory(
                self.memories,
                memory_candidate,
                self.memory_utils.get_memory_prefix(person, self.MODEL_TYPE),
            ):
                memory_to_add = self.memory_utils.add_memory_prefix(
                    memory_candidate, person, self.MODEL_TYPE
                )
                self.memories = self.memory_utils.add_memory(
                    memory_to_add, self.memories
                )
                self.in_session_memories.add(memory_to_add)

        # update mem usage
        used_memory = self_message.get(Module.MEMORY_KNOWLEDGE.message_name(), '')
        self.memories = self.memory_utils.update_memory_usage(
            used_memory, self.memories
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

from parlai.agents.transformer.transformer import add_common_cmdline_args
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_generator_agent import TorchGeneratorAgent
from parlai.utils.logging import logging
from parlai.utils.misc import recursive_getattr
from parlai.utils.torch import padded_tensor

from .modules import (
    PassThroughEncoder,
    TransformerDecoderOnly,
    TransformerGeneratorModel,
)


class DecoderAgent(TorchGeneratorAgent):
    """
    DecoderOnlyAgent.

    Implementation of TorchGeneratorAgent, where the model is a Decoder-Only
    Transformer.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command-line arguments specifically for this agent.
        """
        agent = parser.add_argument_group('Decoder-Only Transformer Arguments')
        add_common_cmdline_args(agent)
        cls.dictionary_class().add_cmdline_args(parser, partial_opt=partial_opt)

        super().add_cmdline_args(parser, partial_opt=partial_opt)
        return agent

    def build_model(self, states=None):
        """
        Override of ``TorchAgent.build_model``.
        """
        assert (
            self.opt['n_encoder_layers'] == -1
        ), "Decoder-only model cannot have encoder layers."
        wrapped_class = TransformerGeneratorModel.with_components(
            encoder=PassThroughEncoder, decoder=TransformerDecoderOnly
        )
        return wrapped_class(self.opt, self.dict)

    def _pad_tensor(self, items, is_label=False):
        """
        Override of ``TorchAgent._pad_tensor``.

        Pads context tensor on the left and label tensor on the right, such that when
        they are concatenated the example meets in the middle to form a continuous
        sequence.
        """
        return padded_tensor(
            items,
            pad_idx=self.NULL_IDX,
            left_padded=(not is_label),
            fp16friendly=self.fp16,
        )

    def _resize_token_embeddings(self, state_dict, msg=None):
        """
        Resize the token embeddings when adding extra special tokens.
        """
        # map extra special tokens carefully
        new_size = self.model.embeddings.weight.size()[0]
        orig_size = state_dict['embeddings.weight'].size()[0]
        logging.info(f'Resizing token embeddings from {orig_size} to {new_size}')
        if new_size <= orig_size:
            # new size should be greater than original size,
            # as we are adding special tokens
            raise RuntimeError(msg)

        for emb_weights in ['embeddings.weight', 'decoder.embeddings.weight']:
            # get new_embs
            old_embs = state_dict[emb_weights]
            new_embs = recursive_getattr(self.model, emb_weights).to(old_embs.device)
            # copy over old weights
            new_embs.data[:orig_size, :] = old_embs.data[:orig_size, :]
            # reset in state dict
            state_dict[emb_weights] = new_embs

        return state_dict



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Imported from https://github.com/hendrycks/math/
Paper: "Measuring Mathematical Problem Solving With the MATH Dataset" (Hendricks et al 2021) https://arxiv.org/pdf/2103.03874.pdf

"""

import glob
import json
import os
import random
import re

from parlai.tasks.math_dataset.build import build
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.io import PathManager
from typing import List, Optional

from parlai.tasks.reasoning.agents import MWPStepsReasoningTeacher


DOMAINS = [
    "algebra",
    "counting_and_probability",
    "geometry",
    "intermediate_algebra",
    "number_theory",
    "prealgebra",
    "precalculus",
]


class MathDatasetStepByStepReasoningTeacher(MWPStepsReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following math problem. Show your work."
        )
        parser.add_argument(
            "--math_dataset-domains",
            nargs="+",
            default=DOMAINS,
            choices=DOMAINS,
        )
        parser.add_argument(
            "--math-latex-converter",
            type=bool,
            default=False,
        )
        parser.add_argument(
            "--extrinsic-step",
            type=bool,
            default=False,
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)  # NOTE: the call to build here
        suffix = 'test' if opt['datatype'].startswith('test') else 'train'
        opt['datafile'] = os.path.join(opt['datapath'], 'MATH', suffix)
        self.id = 'math_dataset'
        self.math_random = random.Random(42)
        super().__init__(opt, shared)

    def load_data(self, domains) -> List[str]:
        data = []
        data_path = self.opt['datafile']
        for domain in domains:
            for filename in glob.glob(f"{data_path}/{domain}/*.json", recursive=True):
                with PathManager.open(filename, "r") as f:
                    example = json.load(f)
                    data.append(example)
        return data

    def get_data_for_fold(self, fold):
        domains = self.opt.get("math_dataset_domains", DOMAINS)
        convert = self.opt.get("math_latex_converter", False)
        extrinsic_step = self.opt.get("extrinsic_step", False)
        data = self.load_data(domains)

        if "test" not in self.datatype:
            # Synthetically split some of train out so we get a valid set.
            # MATHS has 7.5k train, 5k test, 12.5 total.
            split_size = len(data) // 10
            if "valid" in self.datatype:
                data = data[:split_size]
            elif "train" in self.datatype:
                data = data[split_size:]
            else:
                raise RuntimeError(f"Not a viable datatype {self.datatype}!")

        for problem in data:
            question = problem["problem"]
            answer_blob = problem["solution"]
            final_answer = self.get_boxed_answer(answer_blob)
            answer_blob = self._clean_steps(answer_blob)
            steps = answer_blob.split(". ")
            if extrinsic_step:
                random_step = self._find_nonempty_random_step(data)
            if convert:
                question = self._latex_conversion(question)
                final_answer = self._latex_conversion(final_answer)
                final_steps = []
                for step in steps:
                    final_steps.append(self._latex_conversion(step))
                if extrinsic_step:
                    random_step = self._latex_conversion(random_step)
            else:
                final_steps = steps
            while len(final_steps) > 0 and (
                final_steps[-1] == '' or final_steps[-1] == ' '
            ):
                final_steps = final_steps[:-1]
            if extrinsic_step:
                yield {
                    "question": question,
                    "answer": final_answer,
                    "steps": final_steps,
                    "extrinsic_step": random_step,
                }
                continue
            yield {"question": question, "answer": final_answer, "steps": final_steps}

    def _clean_steps(self, steps: str) -> str:
        steps = steps.replace("\n\n", "\n")
        steps = steps.replace("$$", "$ ")
        steps = steps.replace("\n", ". ")
        steps = steps.replace(".$", "$. ")
        steps = steps.replace("..", ".")
        steps = steps.replace(",$", "$,")
        steps = steps.replace(". . ", ". ")

        return steps

    def _latex_conversion(self, final_answer: str) -> str:
        """
        Following page 18 at https://arxiv.org/pdf/2206.14858.pdf with some
        modifications.
        """
        substitutions = [
            ('an ', ''),  # ('a ', ''),
            ('.$', '$'),
            ('\\$', ''),
            (r'\ ', ''),
            # (' ', ''),
            ('  ', ' '),
            ('mbox', 'text'),
            (',\\text{and}', ','),
            ('\\text{and}', ','),
            ('\\text{m}', '\\text{}'),
            ('\\cdot', '*'),
            ('\\left', ''),
            ('\\right', ''),
        ]
        removed_expressions = [
            # 'square', 'ways', 'integers', 'dollars', 'mph', 'inches', 'ft',
            # 'hours', 'km', 'units',
            '\\ldots',
            # 'sue', 'points', 'feet',
            # 'minutes', 'digits', 'cents', 'degrees', 'cm', 'gm', 'pounds',
            # 'meters', 'meals', 'edges', 'students', 'childrentickets', 'multiples',
            '\\text{s}',
            '\\text{.}',
            '\\text{\ns}',
            '\\text{}^2',
            '\\text{}^3',
            '\\text{\n}',
            '\\text{}',
            r'\mathrm{th}',
            r'^\circ',
            r'^{\circ}',
            r'\;',
            r',\!',
            '{,}',
            '"',
            '\\dots',
        ]

        # final_answer = final_answer.split('=')[-1]
        for before, after in substitutions:
            final_answer = final_answer.replace(before, after)
        for expr in removed_expressions:
            final_answer = final_answer.replace(expr, '')

        # Extract answer that is in LaTeX math, is bold,
        # is surrounded by a box, etc.
        # final_answer = re.sub(r'(.*?)(\$)(.*?)(\$)(.*)', '$\\3$', final_answer)
        final_answer = re.sub(r'(\\text\{)(.*?)(\})', '\\2', final_answer)
        final_answer = re.sub(r'(\\textbf\{)(.*?)(\})', '\\2', final_answer)
        final_answer = re.sub(r'(\\overline\{)(.*?)(\})', '\\2', final_answer)
        final_answer = re.sub(r'(\\boxed\{)(.*)(\})', '\\2', final_answer)

        # Normalize shorthand TeX:
        # \fracab -> \frac{a}{b}
        # \frac{abc}{bef} -> \frac{abc}{bef}
        # \fracabc -> \frac{a}{b}c
        # \sqrta -> \sqrt{a}
        # \sqrtab -> sqrt{a}b
        final_answer = re.sub(r'(frac)([^{])(.)', 'frac{\\2}{\\3}', final_answer)
        final_answer = re.sub(r'(sqrt)([^{])', 'sqrt{\\2}', final_answer)
        final_answer = final_answer.replace('$', '')
        # replace \frac{a}{b}
        final_answer = re.sub(
            r'(\\frac){(-?[0-9]*[a-x]*)}{(-?[0-9]*[a-x]*)}', '\\2/\\3', final_answer
        )
        # replace \frac{a}b
        final_answer = re.sub(r'(\\frac){(-?[0-9a-x^()]*)}', '\\2/', final_answer)

        # Normalize 100,000 -> 100000
        if final_answer.replace(',', '').isdigit():
            final_answer = final_answer.replace(',', '')

        return final_answer

    def _find_nonempty_random_step(self, dataset: List[str]) -> str:
        '''Here we *ASSUME* that the whole dataset contains at least one non-empty step
        Otherwise it will go into infinite loop looking for the one
        '''
        # what we call an empty step
        empty_steps = ["", " "]
        # first find chain with at least one non-empty step
        rand_steps = self._clean_steps(
            self.math_random.choice(dataset)["solution"]
        ).split(". ")
        # make sure this chain has at least one non-empty step
        i = 0
        while i < len(rand_steps) and rand_steps[i] in empty_steps:
            i += 1
        # if it doesn't, try again
        if i == len(rand_steps):
            return self._find_nonempty_random_step(dataset)
        random_step = empty_steps[0]
        # find non-empty random step (and we know it exists in this chain)
        while random_step in empty_steps:
            random_step = self.math_random.choice(rand_steps)
        return random_step

    def get_boxed_answer(self, answer):
        boxed_idx = answer.find("boxed{")
        final_answer = answer[boxed_idx:]
        final_answer = final_answer[
            final_answer.find("{") + 1 : final_answer.rfind("}")
        ]
        final_answer = self.clean_answer(final_answer)
        return final_answer

    def clean_answer(self, answer):
        while answer.rfind("}") < answer.rfind('{'):
            answer = answer[: answer.rfind("}")]
        return answer


class DefaultTeacher(MathDatasetStepByStepReasoningTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Imported from https://github.com/chaochun/nlu-asdiv-dataset
Paper: "A Diverse Corpus for Evaluating and Developing English Math Word Problem Solvers" (Miao et al 2021)
https://aclanthology.org/2020.acl-main.92.pdf
"""

import xml.etree.ElementTree as ET
import os

from .build import build
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from typing import Optional

from tasks.reasoning.agents import MWPStepsReasoningTeacher


class ASDivStepByStepReasoningTeacher(MWPStepsReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following math problem. Show your work.",
            thought_token="EQUATION: ",
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)  # NOTE: the call to build here
        opt['datafile'] = os.path.join(opt['datapath'], 'asdiv')
        self.id = 'asdiv'
        super().__init__(opt, shared)

    def load_data(self):
        data = []
        data_path = self.opt['datafile']
        tree = ET.parse(f"{data_path}/ASDiv.xml")
        root = tree.getroot()
        problemset_root = root.getchildren()[0]
        for problem in problemset_root.getchildren():
            problem_elements = problem.getchildren()
            question = problem_elements[0].text + " " + problem_elements[1].text
            answer = problem_elements[3].text
            equation = problem_elements[4].text
            problem_dict = {
                "problem": question,
                "solution": answer,
                "equation": equation,
            }
            data.append(problem_dict)
        return data

    def get_data_for_fold(self, fold):
        data = self.load_data()
        # Synthetically split train/test/valid splits.
        # ASDiv has 2305 problems.
        VALID_SAMPLES = 200
        TEST_SAMPLES = 300
        if "test" not in self.datatype:
            if "valid" in self.datatype:
                data = data[TEST_SAMPLES : TEST_SAMPLES + VALID_SAMPLES]
            elif "train" in self.datatype:
                data = data[TEST_SAMPLES + VALID_SAMPLES :]
            else:
                raise RuntimeError(f"Not a viable datatype {self.datatype}!")
        else:
            data = data[:TEST_SAMPLES]

        for problem in data:
            question = problem["problem"]
            answer = problem["solution"]
            steps = [problem['equation']]
            yield {"question": question, "answer": answer, "steps": steps}


class DefaultTeacher(ASDivStepByStepReasoningTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
SeeKeR Search Query Tasks.
"""
import parlai.tasks.wizard_of_internet.agents as woi
import parlai.utils.logging as logging

import projects.seeker.tasks.mutators  # type: ignore


class WoiSearchQueryTeacher(woi.SearchQueryTeacher):
    def __init__(self, opt, shared=None):
        opt['only_last_search_query'] = True
        mutators = '+'.join(
            ['flatten', 'prompt_search_query_mutator', 'skip_retrieval_mutator']
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
        self.id = 'WoiSearchQueryTeacher'


class SearchQueryTeacher(WoiSearchQueryTeacher):
    pass


class DefaultTeacher(SearchQueryTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from projects.bb3.tasks.r2c2_search_generation_tasks import (
    FitsSearchQueryTeacher as fits_sgm,
    WoiSearchQueryTeacher as woi_sgm,
)


def get_search_generation_opt_mutators(opt: Opt):
    mutators = '+'.join(['prefix_speakers_opt', 'format_gen_tasks_for_decoder_only'])
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class WoiSearchQueryTeacher(woi_sgm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_search_generation_opt_mutators(opt)
        super().__init__(opt, shared)


class FitsSearchQueryTeacher(fits_sgm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_search_generation_opt_mutators(opt)
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Google The Schema-Guided Dialogue(SGD) Dataset implementation for ParlAI.
"""

import os

import parlai.tasks.google_sgd_simulation_splits.build as build_
import parlai.core.tod.tod_core as tod
from parlai.core.metrics import AverageMetric
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
import parlai.core.tod.tod_agents as tod_agents
from parlai.core.opt import Opt
from parlai.tasks.google_sgd.agents import GoogleSGDParser

from typing import List, Optional


class GoogleSgdInDomainParser(GoogleSGDParser):
    """
    Overrides `__init__` and `_load_data` so that we grab our examples from our
    separately constructed custom splits.
    """

    def __init__(self, opt: Opt, shared=None):
        if shared is None:
            # full initialize the teacher as this is not a clone
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        # Best to override here because of init order
        self.dpath = os.path.join(self.opt["datapath"], "google_sgd_rl_splits")
        return super()._load_data(fold)

    def get_id_task_prefix(self):
        return "GoogleSgdInDomain"


class InDomainSystemTeacher(GoogleSgdInDomainParser, tod_agents.TodSystemTeacher):
    pass


class InDomainUserSimulatorTeacher(
    GoogleSgdInDomainParser, tod_agents.TodUserSimulatorTeacher
):
    pass


class InDomainSingleGoalAgent(GoogleSgdInDomainParser, tod_agents.TodSingleGoalAgent):
    pass


class InDomainSingleApiSchemaAgent(
    GoogleSgdInDomainParser, tod_agents.TodSingleApiSchemaAgent
):
    pass


class InDomainGoalAgent(GoogleSgdInDomainParser, tod_agents.TodGoalAgent):
    pass


class InDomainApiSchemaAgent(GoogleSgdInDomainParser, tod_agents.TodApiSchemaAgent):
    pass


class InDomainUserUttAgent(GoogleSgdInDomainParser, tod_agents.TodUserUttAgent):
    pass


class InDomainApiCallAndSysUttAgent(
    GoogleSgdInDomainParser, tod_agents.TodApiCallAndSysUttAgent
):
    pass


VALID_OUT_DOMAIN_API_NAMES = [
    "ShareLocation",
    "RequestPayment",
    "MakePayment",
    "FindApartment",
    "ScheduleVisit",
    "FindHomeByArea",
    "GetCarsAvailable",
    "ReserveCar",
]


class GoogleSgdOutDomainParser(GoogleSGDParser):
    """
    Overrides `__init__` and `_load_data` so that we grab our examples from our
    separately constructed custom splits.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group("Google SGD Out Domain Parser")
        group.add_argument(
            "--filter-single-goal-episodes",
            type=bool,
            default=False,
            help="Filter for only conversations where the original has single goals",
        )

    def __init__(self, opt: Opt, shared=None):
        if shared is None:
            # full initialize the teacher as this is not a clone
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        # Best to override here because of init order
        self.dpath = os.path.join(
            self.opt["datapath"], "google_sgd_rl_splits/model_model_splits"
        )
        return super()._load_data(fold)

    def get_id_task_prefix(self):
        return "GoogleSgdOutDomain"

    def filter_goals(self, goals):
        """
        Used in single goal/api schema agents only.
        """
        result = []
        for goal in goals:
            if goal["api_name"] in VALID_OUT_DOMAIN_API_NAMES:
                result.append(goal)
        return result

    def generate_episodes(self) -> List[tod.TodStructuredEpisode]:
        data = super().generate_episodes()
        if self.opt.get("filter_single_goal_episodes"):
            result = []
            for episode in data:
                if len(episode.goal_calls_machine) == 1:
                    result.append(episode)
            data = result
        return data

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        super().custom_evaluation(teacher_action, labels, model_response)
        resp = model_response.get("text")
        if not resp:
            return

        if (
            teacher_action["type"] == tod.STANDARD_CALL
            and tod.STANDARD_API_NAME_SLOT in teacher_action["slots"]
            and teacher_action["slots"][tod.STANDARD_API_NAME_SLOT]
            in VALID_OUT_DOMAIN_API_NAMES
        ):
            if resp.startswith(tod.STANDARD_CALL):
                resp = resp[len(tod.STANDARD_CALL) :]
            predicted = tod.SerializationHelpers.str_to_api_dict(resp)
            self.metrics.add(
                f"OutDomainOnlyApis/jga",
                AverageMetric(teacher_action["slots"] == predicted),
            )


class OutDomainStandaloneApiTeacher(
    GoogleSgdOutDomainParser, tod_agents.TodStandaloneApiTeacher
):
    pass


class OutDomainSystemTeacher(GoogleSgdOutDomainParser, tod_agents.TodSystemTeacher):
    pass


class OutDomainUserSimulatorTeacher(
    GoogleSgdOutDomainParser, tod_agents.TodUserSimulatorTeacher
):
    pass


class OutDomainGoalAgent(GoogleSgdOutDomainParser, tod_agents.TodGoalAgent):
    pass


class OutDomainApiSchemaAgent(GoogleSgdOutDomainParser, tod_agents.TodApiSchemaAgent):
    pass


class OutDomainSingleGoalAgent(GoogleSgdOutDomainParser, tod_agents.TodSingleGoalAgent):
    pass


class OutDomainSingleApiSchemaAgent(
    GoogleSgdOutDomainParser, tod_agents.TodSingleApiSchemaAgent
):
    pass


class OutDomainUserUttAgent(GoogleSgdOutDomainParser, tod_agents.TodUserUttAgent):
    pass


class OutDomainApiCallAndSysUttAgent(
    GoogleSgdOutDomainParser, tod_agents.TodApiCallAndSysUttAgent
):
    pass


class OutDomainApiResponseAgent(
    GoogleSgdOutDomainParser, tod_agents.TodApiResponseAgent
):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import copy
import os
from .build import build

from parlai.core.teachers import ConversationTeacher


def _path(opt):
    # build the data if it does not exist
    build(opt)

    # set up path to data (specific to each dataset)
    datatype = opt['datatype'].split(':')[0]
    return os.path.join(opt['datapath'], 'LCCC', datatype + '.json')


class LCCCTeacher(ConversationTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('LCCC Task Arguments')
        agent.add_argument(
            '--label-turns',
            type=str,
            help='which speaker to use as label',
            choices=['firstspeaker', 'secondspeaker', 'both'],
            default='secondspeaker',
        )
        return parser

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        # get datafile
        opt['conversationteacher_datafile'] = _path(opt)
        super().__init__(opt, shared)


class DefaultTeacher(LCCCTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from parlai.core.build_data import built, download_models, get_model_dir

PROJECT_NAME = 'multilight'
VERSION = 'v1.0'


def maybe_download(model_type, datapath):
    ddir = os.path.join(get_model_dir(datapath), PROJECT_NAME)
    if not built(os.path.join(ddir, model_type), VERSION):
        opt = {'datapath': datapath, 'model_type': model_type}
        fnames = ['model.tar.gz']
        download_models(opt, fnames, PROJECT_NAME, version=VERSION, use_model_type=True)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Automated Expanded Attention 1024-truncation model.

Uses RPA classifier to select which tokens from the input to re-attend to.
"""
from parlai.zoo.light_whoami.whoami_download import download_with_model_type


def download(datapath):
    download_with_model_type(datapath, 'automated_expanded_attention_1024', 'v1.0')



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestInDomainSystemTeacher(AutoTeacherTest):
    task = "google_sgd_simulation_splits:InDomainSystemTeacher"


class TestInDomainUserSimulatorTeacher(AutoTeacherTest):
    task = "google_sgd_simulation_splits:InDomainUserSimulatorTeacher"


class TestOutDomainSystemTeacher(AutoTeacherTest):
    task = "google_sgd_simulation_splits:OutDomainSystemTeacher"


class TestOutDomainUserSimulatorTeacher(AutoTeacherTest):
    task = "google_sgd_simulation_splits:OutDomainUserSimulatorTeacher"



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Task Oriented Dialogue (TOD) enums and base classes.

This file defines standard tokens, classes for round and conversation structure, and a serialization class to aid in converting between these.

See `tod_agents.py` for usage of these classes to generate training data and `tod_world_script.py` for usage of these classes in simulated conversations.
"""
from enum import Enum
from typing import List, Dict
from dataclasses import dataclass, field
from collections.abc import Iterable
from parlai.utils.misc import warn_once

STANDARD_CALL = "APICALL: "
STANDARD_RESP = "APIRESP: "
STANDARD_SYSTEM_UTTERANCE = "SYSTEM: "
STANDARD_USER_UTTERANCE = "USER: "

STANDARD_GOAL = "GOAL: "
STANDARD_API_SCHEMAS = "APIS: "

STANDARD_API_NAME_SLOT = "api_name"
STANDARD_REQUIRED_KEY = "reqArg"
STANDARD_OPTIONAL_KEY = "optArg"
STANDARD_DONE = "[DONE]"

CONST_SILENCE = "__SILENCE__"


class TodAgentType(str, Enum):
    USER_UTT_AGENT = "user_utt_model"
    API_CALL_AGENT = "api_call_model"
    API_RESP_AGENT = "api_resp_model"
    SYSTEM_UTT_AGENT = "system_utt_model"
    API_SCHEMA_GROUNDING_AGENT = "api_schema_grounding_model"
    GOAL_GROUNDING_AGENT = "goal_grounding_model"


TOD_AGENT_TYPE_TO_PREFIX = {
    TodAgentType.USER_UTT_AGENT: STANDARD_USER_UTTERANCE,
    TodAgentType.API_CALL_AGENT: STANDARD_CALL,
    TodAgentType.API_RESP_AGENT: STANDARD_RESP,
    TodAgentType.SYSTEM_UTT_AGENT: STANDARD_SYSTEM_UTTERANCE,
    TodAgentType.API_SCHEMA_GROUNDING_AGENT: STANDARD_API_SCHEMAS,
    TodAgentType.GOAL_GROUNDING_AGENT: STANDARD_GOAL,
}


@dataclass
class TodStructuredRound:
    """
    Dataclass for rounds.

    After the first (grounding) turn, conversations in the TOD structure are rounds of
       1. User Utterance
       2. System API Call
       3. API Implementation API Response
       4. System Utterance

    This class hold that data.
    """

    # Variables set by those using this class
    user_utt: str = ""
    api_call_machine: Dict = field(
        default_factory=dict
    )  # Hashmap of slot keys and slot values. Note that STANDARD_API_NAME_SLOT (`api_name`) is expected to be one of the keys here when this is nonempty; simulation metrics wonky without
    api_resp_machine: Dict = field(default_factory=dict)
    sys_utt: str = ""
    extras: Dict = field(
        default_factory=dict
    )  # Grab bag for extra data. Not currently referenced in any TOD core code, but a convenient leaky abstraction for passing dataset-specific data between Parser classes and realized agents/teachers.

    # Variables derived by class
    api_call_utt: str = field(init=False)
    api_resp_utt: str = field(init=False)

    def __post_init__(self):
        self.api_call_utt = SerializationHelpers.api_dict_to_str(self.api_call_machine)
        self.api_resp_utt = SerializationHelpers.api_dict_to_str(self.api_resp_machine)
        if (
            len(self.api_call_machine) > 0
            and STANDARD_API_NAME_SLOT not in self.api_call_machine
        ):
            warn_once(
                f"{STANDARD_API_NAME_SLOT} missing when API Call present. This may cause issues for simulation metrics."
            )


@dataclass
class TodStructuredEpisode:
    """
    Dataclass for episode-level data.

    This holds the information for grounding turns (Goal calls, API Schemas), the rounds
    of User/System/API Implementation communications, as well as any extra metadata that
    is useful for the episode.
    """

    # Variables set by those using this class
    delex: bool = False  # Set to true and this class will handle delexicalizing call + response utterances based on API calls and responses exposed to this class.
    domain: str = ""  # self-explanatory
    api_schemas_machine: List[Dict[str, List]] = field(
        default_factory=list
    )  # Expected to be a List of Dicts with the API name, required arguments, and optional arguments (specified by consts at the top of this file) as keys
    goal_calls_machine: List[Dict[str, str]] = field(
        default_factory=list
    )  # Machine-formatted API calls
    rounds: List[TodStructuredRound] = field(default_factory=list)  # self explanatory
    extras: Dict = field(
        default_factory=dict
    )  # Grab bag for extra data. Not currently referenced in any TOD core code, but a convenient leaky abstraction for passing dataset-specific data between Parser classes and realized agents/teachers.

    # Variables derived by class
    api_schemas_utt: str = field(init=False)
    goal_calls_utt: str = field(init=False)

    def __post_init__(self):
        self.api_schemas_utt = SerializationHelpers.list_of_maps_to_str(
            self.api_schemas_machine
        )
        self.goal_calls_machine = [
            call for call in self.goal_calls_machine if len(call) > 0
        ]
        self.goal_calls_utt = SerializationHelpers.list_of_maps_to_str(
            self.goal_calls_machine
        )
        # Add a done turn at the end
        self.rounds.append(TodStructuredRound(user_utt=STANDARD_DONE))
        if self.delex:
            accum_slots = (
                {}
            )  # separate since some slot values change as we go. Use this for delex first
            cum_slots = self.get_all_slots()
            for r in self.rounds:
                accum_slots.update(r.api_call_machine)
                accum_slots.update(r.api_resp_machine)
                r.sys_utt = SerializationHelpers.delex(r.sys_utt, accum_slots)
                r.sys_utt = SerializationHelpers.delex(r.sys_utt, cum_slots)

    def get_all_slots(self):
        result = {}
        for r in self.rounds:
            result.update(r.api_call_machine)
            result.update(r.api_resp_machine)
        return result


class SerializationHelpers:
    @classmethod
    def delex(cls, text, slots):
        delex = text
        for slot, value in slots.items():
            if isinstance(value, str):
                delex = delex.replace(value, f"[{slot}]")
            else:
                for v in value:
                    delex = delex.replace(v, f"[{slot}]")
        return delex

    @classmethod
    def inner_list_join(cls, values):
        if isinstance(values, str):
            return values
        return ", ".join(sorted([str(v).strip() for v in values]))

    @classmethod
    def inner_list_split(cls, s):
        if len(s) < 1:
            return s
        if s[0] == "{":  # for case when we're json serializing a dict
            return s
        split = s.split(", ")
        if len(split) == 1:
            return split[0]
        return set(split)

    @classmethod
    def maybe_inner_list_join(cls, values):
        if type(values) is dict:
            return str(values)
        if (
            isinstance(values, str)
            or isinstance(values, int)
            or isinstance(values, float)
        ):
            return values
        elif isinstance(values, Iterable):
            return SerializationHelpers.inner_list_join(values)
        else:
            raise RuntimeError(
                f"invalid type of argument for maybe_inner_list_join: {values}; type {type(values)}"
            )

    @classmethod
    def api_dict_to_str(cls, apidict):
        """
        Used for API Calls and Responses -> Utterance.
        """

        return " ; ".join(
            f"{k} = {SerializationHelpers.maybe_inner_list_join(v)}"
            for k, v in sorted(apidict.items())
        )

    @classmethod
    def str_to_api_dict(cls, string):
        """
        Used for API Call and Response Utterances -> Dict.
        """
        slot_strs = string.split(" ; ")
        result = {}
        for slot_str in slot_strs:
            if " = " not in slot_str:
                continue
            name, value = slot_str.split(" = ", 1)
            name = name.strip()
            value = SerializationHelpers.inner_list_split(value.strip())
            result[name] = value
        return result

    @classmethod
    def outer_list_join(cls, s):
        return " | ".join(s)

    @classmethod
    def outer_list_split(cls, s):
        return s.split(" | ")

    @classmethod
    def str_to_list_of_maps(cls, s):
        return [
            SerializationHelpers.str_to_api_dict(x)
            for x in SerializationHelpers.outer_list_split(s)
        ]

    @classmethod
    def list_of_maps_to_str(cls, list_of_maps):
        return SerializationHelpers.outer_list_join(
            [SerializationHelpers.api_dict_to_str(m) for m in list_of_maps]
        )

    @classmethod
    def str_to_goals(cls, s):  # convenience
        return SerializationHelpers.str_to_list_of_maps(s)

    @classmethod
    def str_to_api_schemas(cls, s):  # convenience
        return SerializationHelpers.str_to_list_of_maps(s)
